# This file contains modules common to various models

import math

import numpy as np
import torch
from torch import nn

from nataili.util.codeformer.facelib.detection.yolov5face.utils.datasets import letterbox
from nataili.util.codeformer.facelib.detection.yolov5face.utils.general import (
    make_divisible,
    non_max_suppression,
    scale_coords,
    xyxy2xywh,
)


def autopad(k, p=None):  # kernel, padding
    # Pad to 'same'
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p


def channel_shuffle(x, groups):
    batchsize, num_channels, height, width = x.data.size()
    channels_per_group = torch.div(num_channels, groups, rounding_mode="trunc")

    # reshape
    x = x.view(batchsize, groups, channels_per_group, height, width)
    x = torch.transpose(x, 1, 2).contiguous()

    # flatten
    return x.view(batchsize, -1, height, width)


def DWConv(c1, c2, k=1, s=1, act=True):
    # Depthwise convolution
    return Conv(c1, c2, k, s, g=math.gcd(c1, c2), act=act)


class Conv(nn.Module):
    # Standard convolution
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g, bias=False)
        self.bn = nn.BatchNorm2d(c2)
        self.act = nn.SiLU() if act is True else (act if isinstance(act, nn.Module) else nn.Identity())

    def forward(self, x):
        return self.act(self.bn(self.conv(x)))

    def fuseforward(self, x):
        return self.act(self.conv(x))


class StemBlock(nn.Module):
    def __init__(self, c1, c2, k=3, s=2, p=None, g=1, act=True):
        super().__init__()
        self.stem_1 = Conv(c1, c2, k, s, p, g, act)
        self.stem_2a = Conv(c2, c2 // 2, 1, 1, 0)
        self.stem_2b = Conv(c2 // 2, c2, 3, 2, 1)
        self.stem_2p = nn.MaxPool2d(kernel_size=2, stride=2, ceil_mode=True)
        self.stem_3 = Conv(c2 * 2, c2, 1, 1, 0)

    def forward(self, x):
        stem_1_out = self.stem_1(x)
        stem_2a_out = self.stem_2a(stem_1_out)
        stem_2b_out = self.stem_2b(stem_2a_out)
        stem_2p_out = self.stem_2p(stem_1_out)
        return self.stem_3(torch.cat((stem_2b_out, stem_2p_out), 1))


class Bottleneck(nn.Module):
    # Standard bottleneck
    def __init__(self, c1, c2, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_, c2, 3, 1, g=g)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))


class BottleneckCSP(nn.Module):
    # CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = nn.Conv2d(c1, c_, 1, 1, bias=False)
        self.cv3 = nn.Conv2d(c_, c_, 1, 1, bias=False)
        self.cv4 = Conv(2 * c_, c2, 1, 1)
        self.bn = nn.BatchNorm2d(2 * c_)  # applied to cat(cv2, cv3)
        self.act = nn.LeakyReLU(0.1, inplace=True)
        self.m = nn.Sequential(*(Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)))

    def forward(self, x):
        y1 = self.cv3(self.m(self.cv1(x)))
        y2 = self.cv2(x)
        return self.cv4(self.act(self.bn(torch.cat((y1, y2), dim=1))))


class C3(nn.Module):
    # CSP Bottleneck with 3 convolutions
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c1, c_, 1, 1)
        self.cv3 = Conv(2 * c_, c2, 1)  # act=FReLU(c2)
        self.m = nn.Sequential(*(Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)))

    def forward(self, x):
        return self.cv3(torch.cat((self.m(self.cv1(x)), self.cv2(x)), dim=1))


class ShuffleV2Block(nn.Module):
    def __init__(self, inp, oup, stride):
        super().__init__()

        if not 1 <= stride <= 3:
            raise ValueError("illegal stride value")
        self.stride = stride

        branch_features = oup // 2

        if self.stride > 1:
            self.branch1 = nn.Sequential(
                self.depthwise_conv(inp, inp, kernel_size=3, stride=self.stride, padding=1),
                nn.BatchNorm2d(inp),
                nn.Conv2d(inp, branch_features, kernel_size=1, stride=1, padding=0, bias=False),
                nn.BatchNorm2d(branch_features),
                nn.SiLU(),
            )
        else:
            self.branch1 = nn.Sequential()

        self.branch2 = nn.Sequential(
            nn.Conv2d(
                inp if (self.stride > 1) else branch_features,
                branch_features,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.BatchNorm2d(branch_features),
            nn.SiLU(),
            self.depthwise_conv(
                branch_features,
                branch_features,
                kernel_size=3,
                stride=self.stride,
                padding=1,
            ),
            nn.BatchNorm2d(branch_features),
            nn.Conv2d(
                branch_features,
                branch_features,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.BatchNorm2d(branch_features),
            nn.SiLU(),
        )

    @staticmethod
    def depthwise_conv(i, o, kernel_size, stride=1, padding=0, bias=False):
        return nn.Conv2d(i, o, kernel_size, stride, padding, bias=bias, groups=i)

    def forward(self, x):
        if self.stride == 1:
            x1, x2 = x.chunk(2, dim=1)
            out = torch.cat((x1, self.branch2(x2)), dim=1)
        else:
            out = torch.cat((self.branch1(x), self.branch2(x)), dim=1)
        out = channel_shuffle(out, 2)
        return out


class SPP(nn.Module):
    # Spatial pyramid pooling layer used in YOLOv3-SPP
    def __init__(self, c1, c2, k=(5, 9, 13)):
        super().__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_ * (len(k) + 1), c2, 1, 1)
        self.m = nn.ModuleList([nn.MaxPool2d(kernel_size=x, stride=1, padding=x // 2) for x in k])

    def forward(self, x):
        x = self.cv1(x)
        return self.cv2(torch.cat([x] + [m(x) for m in self.m], 1))


class Focus(nn.Module):
    # Focus wh information into c-space
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.conv = Conv(c1 * 4, c2, k, s, p, g, act)

    def forward(self, x):  # x(b,c,w,h) -> y(b,4c,w/2,h/2)
        return self.conv(
            torch.cat(
                [
                    x[..., ::2, ::2],
                    x[..., 1::2, ::2],
                    x[..., ::2, 1::2],
                    x[..., 1::2, 1::2],
                ],
                1,
            )
        )


class Concat(nn.Module):
    # Concatenate a list of tensors along dimension
    def __init__(self, dimension=1):
        super().__init__()
        self.d = dimension

    def forward(self, x):
        return torch.cat(x, self.d)


class NMS(nn.Module):
    # Non-Maximum Suppression (NMS) module
    conf = 0.25  # confidence threshold
    iou = 0.45  # IoU threshold
    classes = None  # (optional list) filter by class

    def forward(self, x):
        return non_max_suppression(x[0], conf_thres=self.conf, iou_thres=self.iou, classes=self.classes)


class AutoShape(nn.Module):
    # input-robust model wrapper for passing cv2/np/PIL/torch inputs. Includes preprocessing, inference and NMS
    img_size = 640  # inference size (pixels)
    conf = 0.25  # NMS confidence threshold
    iou = 0.45  # NMS IoU threshold
    classes = None  # (optional list) filter by class

    def __init__(self, model):
        super().__init__()
        self.model = model.eval()

    def autoshape(self):
        print("autoShape already enabled, skipping... ")  # model already converted to model.autoshape()
        return self

    def forward(self, imgs, size=640, augment=False, profile=False):
        # Inference from various sources. For height=720, width=1280, RGB images example inputs are:
        #   OpenCV:          = cv2.imread('image.jpg')[:,:,::-1]  # HWC BGR to RGB x(720,1280,3)
        #   PIL:             = Image.open('image.jpg')  # HWC x(720,1280,3)
        #   numpy:           = np.zeros((720,1280,3))  # HWC
        #   torch:           = torch.zeros(16,3,720,1280)  # BCHW
        #   multiple:        = [Image.open('image1.jpg'), Image.open('image2.jpg'), ...]  # list of images

        p = next(self.model.parameters())  # for device and type
        if isinstance(imgs, torch.Tensor):  # torch
            return self.model(imgs.to(p.device).type_as(p), augment, profile)  # inference

        # Pre-process
        n, imgs = (len(imgs), imgs) if isinstance(imgs, list) else (1, [imgs])  # number of images, list of images
        shape0, shape1 = [], []  # image and inference shapes
        for i, im in enumerate(imgs):
            im = np.array(im)  # to numpy
            if im.shape[0] < 5:  # image in CHW
                im = im.transpose((1, 2, 0))  # reverse dataloader .transpose(2, 0, 1)
            im = im[:, :, :3] if im.ndim == 3 else np.tile(im[:, :, None], 3)  # enforce 3ch input
            s = im.shape[:2]  # HWC
            shape0.append(s)  # image shape
            g = size / max(s)  # gain
            shape1.append([y * g for y in s])
            imgs[i] = im  # update
        shape1 = [make_divisible(x, int(self.stride.max())) for x in np.stack(shape1, 0).max(0)]  # inference shape
        x = [letterbox(im, new_shape=shape1, auto=False)[0] for im in imgs]  # pad
        x = np.stack(x, 0) if n > 1 else x[0][None]  # stack
        x = np.ascontiguousarray(x.transpose((0, 3, 1, 2)))  # BHWC to BCHW
        x = torch.from_numpy(x).to(p.device).type_as(p) / 255.0  # uint8 to fp16/32

        # Inference
        with torch.no_grad():
            y = self.model(x, augment, profile)[0]  # forward
        y = non_max_suppression(y, conf_thres=self.conf, iou_thres=self.iou, classes=self.classes)  # NMS

        # Post-process
        for i in range(n):
            scale_coords(shape1, y[i][:, :4], shape0[i])

        return Detections(imgs, y, self.names)


class Detections:
    # detections class for YOLOv5 inference results
    def __init__(self, imgs, pred, names=None):
        super().__init__()
        d = pred[0].device  # device
        gn = [
            torch.tensor([*(im.shape[i] for i in [1, 0, 1, 0]), 1.0, 1.0], device=d) for im in imgs
        ]  # normalizations
        self.imgs = imgs  # list of images as numpy arrays
        self.pred = pred  # list of tensors pred[0] = (xyxy, conf, cls)
        self.names = names  # class names
        self.xyxy = pred  # xyxy pixels
        self.xywh = [xyxy2xywh(x) for x in pred]  # xywh pixels
        self.xyxyn = [x / g for x, g in zip(self.xyxy, gn)]  # xyxy normalized
        self.xywhn = [x / g for x, g in zip(self.xywh, gn)]  # xywh normalized
        self.n = len(self.pred)

    def __len__(self):
        return self.n

    def tolist(self):
        # return a list of Detections objects, i.e. 'for result in results.tolist():'
        x = [Detections([self.imgs[i]], [self.pred[i]], self.names) for i in range(self.n)]
        for d in x:
            for k in ["imgs", "pred", "xyxy", "xyxyn", "xywh", "xywhn"]:
                setattr(d, k, getattr(d, k)[0])  # pop out of list
        return x



"""
Copyright [2022] Victor C Hall

Licensed under the GNU Affero General Public License;
You may not use this code except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math

import numpy as np
from torch.utils.data import Dataset

from . import dl_singleton as dls
from .data_loader import DataLoaderMultiAspect as dlma
from .image_train_item import ImageTrainItem


class EDValidateBatch(Dataset):
    def __init__(
        self,
        data_root,
        flip_p=0.0,
        repeats=1,
        debug_level=0,
        batch_size=1,
        set="val",
    ):
        self.data_root = data_root
        self.batch_size = batch_size

        if not dls.shared_dataloader:
            print("Creating new dataloader singleton")
            dls.shared_dataloader = dlma(
                data_root=data_root, debug_level=debug_level, batch_size=self.batch_size, flip_p=flip_p
            )

        self.image_train_items = dls.shared_dataloader.get_all_images()

        self.num_images = len(self.image_train_items)

        self._length = max(math.trunc(self.num_images * repeats), batch_size) - self.num_images % self.batch_size

        print()
        print(f" ** Validation Set: {set}, steps: {self._length / batch_size:.0f}, repeats: {repeats} ")
        print()

    def __len__(self):
        return self._length

    def __getitem__(self, i):
        idx = i % self.num_images
        image_train_item = self.image_train_items[idx]

        example = self.__get_image_for_trainer(image_train_item)
        return example

    @staticmethod
    def __get_image_for_trainer(image_train_item: ImageTrainItem):
        example = {}

        image_train_tmp = image_train_item.hydrate()

        example["image"] = image_train_tmp.image
        example["caption"] = image_train_tmp.caption

        return example



"""
Copyright [2022] Victor C Hall

Licensed under the GNU Affero General Public License;
You may not use this code except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import math
import os
import random

import numpy as np
import PIL
from torchvision import transforms, utils

_RANDOM_TRIM = 0.04


class ImageTrainItem:
    """
    image: PIL.Image
    identifier: caption,
    target_aspect: (width, height),
    pathname: path to image file
    flip_p: probability of flipping image (0.0 to 1.0)
    """

    def __init__(self, image: PIL.Image, caption: str, target_wh: list, pathname: str, flip_p=0.0):
        self.caption = caption
        self.target_wh = target_wh
        self.pathname = pathname
        self.flip = transforms.RandomHorizontalFlip(p=flip_p)
        self.cropped_img = None
        self.runt_size = 0

        if image is None:
            self.image = []
        else:
            self.image = image

    def hydrate(self, crop=False, save=False, crop_jitter=20):
        """
        crop: hard center crop to 512x512
        save: save the cropped image to disk, for manual inspection of resize/crop
        crop_jitter: randomly shift cropp by N pixels when using multiple aspect ratios to improve training quality
        """
        # print(self.pathname, self.image)
        try:
            # if not hasattr(self, 'image'):
            self.image = PIL.Image.open(self.pathname).convert("RGB")

            width, height = self.image.size
            if crop:
                cropped_img = self.__autocrop(self.image)
                self.image = cropped_img.resize((512, 512), resample=PIL.Image.BICUBIC)
            else:
                width, height = self.image.size
                jitter_amount = random.randint(0, crop_jitter)

                if self.target_wh[0] == self.target_wh[1]:
                    if width > height:
                        left = random.randint(0, width - height)
                        self.image = self.image.crop((left, 0, height + left, height))
                        width = height
                    elif height > width:
                        top = random.randint(0, height - width)
                        self.image = self.image.crop((0, top, width, width + top))
                        height = width
                    elif width > self.target_wh[0]:
                        slice = min(int(self.target_wh[0] * _RANDOM_TRIM), width - self.target_wh[0])
                        slicew_ratio = random.random()
                        left = int(slice * slicew_ratio)
                        right = width - int(slice * (1 - slicew_ratio))
                        sliceh_ratio = random.random()
                        top = int(slice * sliceh_ratio)
                        bottom = height - int(slice * (1 - sliceh_ratio))

                        self.image = self.image.crop((left, top, right, bottom))
                else:
                    image_aspect = width / height
                    target_aspect = self.target_wh[0] / self.target_wh[1]
                    if image_aspect > target_aspect:
                        new_width = int(height * target_aspect)
                        jitter_amount = max(min(jitter_amount, int(abs(width - new_width) / 2)), 0)
                        left = jitter_amount
                        right = left + new_width
                        self.image = self.image.crop((left, 0, right, height))
                    else:
                        new_height = int(width / target_aspect)
                        jitter_amount = max(min(jitter_amount, int(abs(height - new_height) / 2)), 0)
                        top = jitter_amount
                        bottom = top + new_height
                        self.image = self.image.crop((0, top, width, bottom))
                self.image = self.image.resize(self.target_wh, resample=PIL.Image.BICUBIC)

            self.image = self.flip(self.image)
        except Exception as e:
            logging.error(f"Fatal Error loading image: {self.pathname}:")
            logging.error(e)
            exit()

        if type(self.image) is not np.ndarray:
            if save:
                base_name = os.path.basename(self.pathname)
                if not os.path.exists("test/output"):
                    os.makedirs("test/output")
                self.image.save(f"test/output/{base_name}")

            self.image = np.array(self.image).astype(np.uint8)

            # self.image = (self.image / 127.5 - 1.0).astype(np.float32)

        # print(self.image.shape)

        return self

    @staticmethod
    def __autocrop(image: PIL.Image, q=0.404):
        """
        crops image to a random square inside small axis using a truncated gaussian distribution across the long axis
        """
        x, y = image.size

        if x != y:
            if x > y:
                rand_x = x - y
                sigma = max(rand_x * q, 1)
            else:
                rand_y = y - x
                sigma = max(rand_y * q, 1)

            if x > y:
                x_crop_gauss = abs(random.gauss(0, sigma))
                x_crop = min(x_crop_gauss, (x - y) / 2)
                x_crop = math.trunc(x_crop)
                y_crop = 0
            else:
                y_crop_gauss = abs(random.gauss(0, sigma))
                x_crop = 0
                y_crop = min(y_crop_gauss, (y - x) / 2)
                y_crop = math.trunc(y_crop)

            min_xy = min(x, y)
            image = image.crop((x_crop, y_crop, x_crop + min_xy, y_crop + min_xy))

        return image



import numpy as np
from numpy.linalg import inv, lstsq
from numpy.linalg import matrix_rank as rank
from numpy.linalg import norm


class MatlabCp2tormException(Exception):
    def __str__(self):
        return "In File {}:{}".format(__file__, super.__str__(self))


def tformfwd(trans, uv):
    """
    Function:
    ----------
        apply affine transform 'trans' to uv

    Parameters:
    ----------
        @trans: 3x3 np.array
            transform matrix
        @uv: Kx2 np.array
            each row is a pair of coordinates (x, y)

    Returns:
    ----------
        @xy: Kx2 np.array
            each row is a pair of transformed coordinates (x, y)
    """
    uv = np.hstack((uv, np.ones((uv.shape[0], 1))))
    xy = np.dot(uv, trans)
    xy = xy[:, 0:-1]
    return xy


def tforminv(trans, uv):
    """
    Function:
    ----------
        apply the inverse of affine transform 'trans' to uv

    Parameters:
    ----------
        @trans: 3x3 np.array
            transform matrix
        @uv: Kx2 np.array
            each row is a pair of coordinates (x, y)

    Returns:
    ----------
        @xy: Kx2 np.array
            each row is a pair of inverse-transformed coordinates (x, y)
    """
    Tinv = inv(trans)
    xy = tformfwd(Tinv, uv)
    return xy


def findNonreflectiveSimilarity(uv, xy, options=None):
    options = {"K": 2}

    K = options["K"]
    M = xy.shape[0]
    x = xy[:, 0].reshape((-1, 1))  # use reshape to keep a column vector
    y = xy[:, 1].reshape((-1, 1))  # use reshape to keep a column vector

    tmp1 = np.hstack((x, y, np.ones((M, 1)), np.zeros((M, 1))))
    tmp2 = np.hstack((y, -x, np.zeros((M, 1)), np.ones((M, 1))))
    X = np.vstack((tmp1, tmp2))

    u = uv[:, 0].reshape((-1, 1))  # use reshape to keep a column vector
    v = uv[:, 1].reshape((-1, 1))  # use reshape to keep a column vector
    U = np.vstack((u, v))

    # We know that X * r = U
    if rank(X) >= 2 * K:
        r, _, _, _ = lstsq(X, U, rcond=-1)
        r = np.squeeze(r)
    else:
        raise Exception("cp2tform:twoUniquePointsReq")
    sc = r[0]
    ss = r[1]
    tx = r[2]
    ty = r[3]

    Tinv = np.array([[sc, -ss, 0], [ss, sc, 0], [tx, ty, 1]])
    T = inv(Tinv)
    T[:, 2] = np.array([0, 0, 1])

    return T, Tinv


def findSimilarity(uv, xy, options=None):
    options = {"K": 2}

    #    uv = np.array(uv)
    #    xy = np.array(xy)

    # Solve for trans1
    trans1, trans1_inv = findNonreflectiveSimilarity(uv, xy, options)

    # Solve for trans2

    # manually reflect the xy data across the Y-axis
    xyR = xy
    xyR[:, 0] = -1 * xyR[:, 0]

    trans2r, trans2r_inv = findNonreflectiveSimilarity(uv, xyR, options)

    # manually reflect the tform to undo the reflection done on xyR
    TreflectY = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, 1]])

    trans2 = np.dot(trans2r, TreflectY)

    # Figure out if trans1 or trans2 is better
    xy1 = tformfwd(trans1, uv)
    norm1 = norm(xy1 - xy)

    xy2 = tformfwd(trans2, uv)
    norm2 = norm(xy2 - xy)

    if norm1 <= norm2:
        return trans1, trans1_inv
    else:
        trans2_inv = inv(trans2)
        return trans2, trans2_inv


def get_similarity_transform(src_pts, dst_pts, reflective=True):
    """
    Function:
    ----------
        Find Similarity Transform Matrix 'trans':
            u = src_pts[:, 0]
            v = src_pts[:, 1]
            x = dst_pts[:, 0]
            y = dst_pts[:, 1]
            [x, y, 1] = [u, v, 1] * trans

    Parameters:
    ----------
        @src_pts: Kx2 np.array
            source points, each row is a pair of coordinates (x, y)
        @dst_pts: Kx2 np.array
            destination points, each row is a pair of transformed
            coordinates (x, y)
        @reflective: True or False
            if True:
                use reflective similarity transform
            else:
                use non-reflective similarity transform

    Returns:
    ----------
       @trans: 3x3 np.array
            transform matrix from uv to xy
        trans_inv: 3x3 np.array
            inverse of trans, transform matrix from xy to uv
    """

    if reflective:
        trans, trans_inv = findSimilarity(src_pts, dst_pts)
    else:
        trans, trans_inv = findNonreflectiveSimilarity(src_pts, dst_pts)

    return trans, trans_inv


def cvt_tform_mat_for_cv2(trans):
    """
    Function:
    ----------
        Convert Transform Matrix 'trans' into 'cv2_trans' which could be
        directly used by cv2.warpAffine():
            u = src_pts[:, 0]
            v = src_pts[:, 1]
            x = dst_pts[:, 0]
            y = dst_pts[:, 1]
            [x, y].T = cv_trans * [u, v, 1].T

    Parameters:
    ----------
        @trans: 3x3 np.array
            transform matrix from uv to xy

    Returns:
    ----------
        @cv2_trans: 2x3 np.array
            transform matrix from src_pts to dst_pts, could be directly used
            for cv2.warpAffine()
    """
    cv2_trans = trans[:, 0:2].T

    return cv2_trans


def get_similarity_transform_for_cv2(src_pts, dst_pts, reflective=True):
    """
    Function:
    ----------
        Find Similarity Transform Matrix 'cv2_trans' which could be
        directly used by cv2.warpAffine():
            u = src_pts[:, 0]
            v = src_pts[:, 1]
            x = dst_pts[:, 0]
            y = dst_pts[:, 1]
            [x, y].T = cv_trans * [u, v, 1].T

    Parameters:
    ----------
        @src_pts: Kx2 np.array
            source points, each row is a pair of coordinates (x, y)
        @dst_pts: Kx2 np.array
            destination points, each row is a pair of transformed
            coordinates (x, y)
        reflective: True or False
            if True:
                use reflective similarity transform
            else:
                use non-reflective similarity transform

    Returns:
    ----------
        @cv2_trans: 2x3 np.array
            transform matrix from src_pts to dst_pts, could be directly used
            for cv2.warpAffine()
    """
    trans, trans_inv = get_similarity_transform(src_pts, dst_pts, reflective)
    cv2_trans = cvt_tform_mat_for_cv2(trans)

    return cv2_trans


if __name__ == "__main__":
    """
    u = [0, 6, -2]
    v = [0, 3, 5]
    x = [-1, 0, 4]
    y = [-1, -10, 4]

    # In Matlab, run:
    #
    #   uv = [u'; v'];
    #   xy = [x'; y'];
    #   tform_sim=cp2tform(uv,xy,'similarity');
    #
    #   trans = tform_sim.tdata.T
    #   ans =
    #       -0.0764   -1.6190         0
    #        1.6190   -0.0764         0
    #       -3.2156    0.0290    1.0000
    #   trans_inv = tform_sim.tdata.Tinv
    #    ans =
    #
    #       -0.0291    0.6163         0
    #       -0.6163   -0.0291         0
    #       -0.0756    1.9826    1.0000
    #    xy_m=tformfwd(tform_sim, u,v)
    #
    #    xy_m =
    #
    #       -3.2156    0.0290
    #        1.1833   -9.9143
    #        5.0323    2.8853
    #    uv_m=tforminv(tform_sim, x,y)
    #
    #    uv_m =
    #
    #        0.5698    1.3953
    #        6.0872    2.2733
    #       -2.6570    4.3314
    """
    u = [0, 6, -2]
    v = [0, 3, 5]
    x = [-1, 0, 4]
    y = [-1, -10, 4]

    uv = np.array((u, v)).T
    xy = np.array((x, y)).T

    print("\n--->uv:")
    print(uv)
    print("\n--->xy:")
    print(xy)

    trans, trans_inv = get_similarity_transform(uv, xy)

    print("\n--->trans matrix:")
    print(trans)

    print("\n--->trans_inv matrix:")
    print(trans_inv)

    print("\n---> apply transform to uv")
    print("\nxy_m = uv_augmented * trans")
    uv_aug = np.hstack((uv, np.ones((uv.shape[0], 1))))
    xy_m = np.dot(uv_aug, trans)
    print(xy_m)

    print("\nxy_m = tformfwd(trans, uv)")
    xy_m = tformfwd(trans, uv)
    print(xy_m)

    print("\n---> apply inverse transform to xy")
    print("\nuv_m = xy_augmented * trans_inv")
    xy_aug = np.hstack((xy, np.ones((xy.shape[0], 1))))
    uv_m = np.dot(xy_aug, trans_inv)
    print(uv_m)

    print("\nuv_m = tformfwd(trans_inv, xy)")
    uv_m = tformfwd(trans_inv, xy)
    print(uv_m)

    uv_m = tforminv(trans, xy)
    print("\nuv_m = tforminv(trans, xy)")
    print(uv_m)



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import re
import time
from contextlib import nullcontext
from typing import List, Literal, Optional, Tuple, Union

import einops
import k_diffusion as K
import numpy as np
import skimage
import torch
from einops import rearrange
from PIL import Image, ImageOps
from slugify import slugify
from torch import nn
from transformers import CLIPFeatureExtractor

from annotator.canny import CannyDetector
from annotator.util import HWC3
from annotator.util import resize_image as control_resize_image
from cldm.cldm import ControlLDM
from ldm2.models.diffusion.dpm_solver import DPMSolverSampler
from ldm.models.diffusion.ddim import DDIMSampler
from ldm.models.diffusion.ddpm import LatentDiffusion
from ldm.models.diffusion.ddpm_edit import LatentDiffusion as LatentDiffusionPix2Pix
from ldm.models.diffusion.kdiffusion import CFGMaskedDenoiser, KDiffusionSampler
from ldm.models.diffusion.plms import PLMSSampler
from nataili import InvalidModelCacheException
from nataili.model_manager.controlnet import ControlNetModelManager
from nataili.stable_diffusion.annotation import (
    HED,
    Canny,
    Depth,
    FakeScribbles,
    Hough,
    Normal,
    Openpose,
    Scribble,
    Seg,
)
from nataili.stable_diffusion.prompt_weights import (
    fix_mismatched_tensors,
    get_learned_conditioning_with_prompt_weights,
)
from nataili.util.cast import autocast_cpu, autocast_cuda
from nataili.util.create_random_tensors import create_random_tensors
from nataili.util.get_next_sequence_number import get_next_sequence_number
from nataili.util.img2img import find_noise_for_image, get_matched_noise, process_init_mask, resize_image
from nataili.util.logger import logger
from nataili.util.process_prompt_tokens import process_prompt_tokens
from nataili.util.save_sample import save_sample
from nataili.util.seed_to_int import seed_to_int

try:
    from nataili.util.voodoo import load_from_plasma
except ModuleNotFoundError as e:
    from nataili import disable_voodoo

    if not disable_voodoo.active:
        raise e


def offload_model(model: nn.Module, cpu=None, gpu=None):
    if cpu is None and gpu is None:
        raise ValueError("Must specify either cpu or gpu")
    if cpu and gpu:
        raise ValueError("Cannot offload model to both cpu and gpu")
    if cpu:
        model = model.cpu()
    if gpu:
        model = model.to(gpu)
    return model


def low_vram_mode():
    """
    enabled by default
    """
    return os.environ.get("LOW_VRAM_MODE", "1") == "1"


"""
has no effect if low_vram_mode is disabled or if disable_low_vram_mode is enabled
NOTE:
"cpu" and "cuda" are the supported device types
"cpu" could be replaced with "offload", the check would need to account for that
the check is "cuda" in device, so "cuda:0" would also be supported
other device types can be added as needed "mps" etc -
    * add a check for the device type
    * pass that device type in the list of models

list of models is a list of tuples (model, device)
example:
    models = [(model, "cpu"), (model.cond_stage_model.transformer, "cuda:0")]
"""


def low_vram(models: list, force: bool = False):
    if not low_vram_mode() and not force:
        return
    for m in models:
        model, device = m
        if not isinstance(model, nn.Module):
            raise ValueError("model must be a torch.nn.Module")
        if type(device) == torch.device:
            model = offload_model(model, gpu=device)
        elif type(device) == str:
            if device == "cpu":
                model = offload_model(model, cpu=device)
            elif "cuda" in device:
                model = offload_model(model, gpu=device)
            else:
                raise ValueError("Unknown device")
        else:
            raise ValueError(f"Unknown device type {type(device)}")


class CompVis:
    def __init__(
        self,
        model,
        output_dir,
        model_name=None,
        model_baseline=None,
        save_extension="jpg",
        output_file_path=False,
        load_concepts=False,
        concepts_dir=None,
        filter_nsfw=False,
        safety_checker=None,
        disable_voodoo=False,
        control_net_manager=None,
    ):
        self.model = model
        self.model_name = model_name
        self.model_baseline = model_baseline
        self.output_dir = output_dir
        self.output_file_path = output_file_path
        self.save_extension = save_extension
        self.load_concepts = load_concepts
        self.concepts_dir = concepts_dir
        self.output_images = []
        self.images = []
        self.filter_nsfw = filter_nsfw
        self.safety_checker = safety_checker
        self.feature_extractor = CLIPFeatureExtractor()
        self.disable_voodoo = disable_voodoo
        self.control_net_manager: ControlNetModelManager = (
            ControlNetModelManager() if control_net_manager is None else control_net_manager
        )
        self.control_net_model: ControlLDM = None

    @autocast_cuda
    def generate(
        self,
        prompt: str,
        init_img: Image.Image = None,
        init_mask: Image.Image = None,
        mask_mode: Literal["mask", "invert", "alpha"] = "mask",
        resize_mode: Literal["resize", "crop", "resize_and_crop"] = "resize",
        noise_mode: Literal["seed", "find", "matched", "find_and_matched"] = "seed",
        find_noise_steps: int = 50,
        denoising_strength: float = 0.8,
        ddim_steps: int = 50,
        sampler_name="k_lms",
        n_iter: int = 1,
        batch_size: int = 1,
        cfg_scale: float = 7.5,
        seed: Union[int, str] = None,
        height: int = 512,
        width: int = 512,
        save_individual_images: bool = True,
        save_grid: bool = True,
        ddim_eta: float = 0.0,
        sigma_override: dict = None,
        tiling: bool = False,
        clip_skip: int = 1,
        hires_fix: bool = False,
        control_type: Literal[
            "canny", "hed", "depth", "normal", "openpose", "seg", "scribble", "fakescribbles", "hough"
        ] = None,
        init_as_control: bool = False,
        return_control_map: bool = False,
        inpainting: bool = False,
    ):
        model_context = (
            load_from_plasma(self.model["model"], self.model["device"]) if not self.disable_voodoo else nullcontext()
        )
        with model_context as model:
            if self.disable_voodoo:
                model: Union[LatentDiffusion, LatentDiffusionPix2Pix, ControlLDM] = self.model["model"]
            if (
                control_type is not None
                and (
                    ("stable diffusion 2" in self.model_baseline and control_type not in ["normal", "mlsd", "hough"])
                    or ("stable diffusion 2" not in self.model_baseline)
                )
                and init_img is not None
                and self.model_name != "pix2pix"
            ):
                sampler_name = "DDIM"
                init_img = resize_image(resize_mode, init_img, width, height)
                low_vram(
                    [
                        (model, "cpu"),
                        (
                            model.cond_stage_model.transformer
                            if hasattr(model.cond_stage_model, "transformer")
                            else model.cond_stage_model.model.transformer,
                            "cpu",
                        ),
                        (model.first_stage_model, "cpu"),
                    ],
                    force=True,  # always move original model to cpu
                )
                model.cond_stage_model.device = "cpu"
                if self.model_baseline == "stable diffusion 2":
                    model_suffix = f"{control_type}_sd2"
                else:
                    model_suffix = control_type
                resolution = round(min(height, width) / 64) * 64
                self.control_net_manager.load_controlnet(f"control_{model_suffix}")
                self.control_net_manager.load_control_ldm(
                    f"control_{model_suffix}", self.model_name, model.state_dict(), _device=self.model["device"]
                )
                loaded_control_ldm = f"control_{model_suffix}_{self.model_name}"
                self.model_name = loaded_control_ldm
                self.control_net_model = self.control_net_manager.loaded_models[loaded_control_ldm]["model"]
                if control_type == "canny":
                    result = Canny(init_as_control)(init_img, resolution=resolution)
                    control = result["control"]
                elif control_type == "hed":
                    result = HED(init_as_control)(init_img, resolution=resolution, detect_resolution=resolution)
                    control = result["control"]
                elif control_type == "depth":
                    result = Depth(init_as_control)(init_img, resolution=resolution)
                    control = result["control"]
                elif control_type == "scribble":
                    # init can't be used as control for scribbles
                    result = Scribble()(init_img, resolution=resolution)
                    control = result["control"]
                elif control_type == "fakescribbles":
                    result = FakeScribbles(init_as_control)(
                        init_img, resolution=resolution, detect_resolution=resolution
                    )
                    control = result["control"]
                elif control_type == "hough":
                    result = Hough(init_as_control)(init_img, resolution=resolution, detect_resolution=resolution)
                    control = result["control"]
                elif control_type == "openpose":
                    result = Openpose(init_as_control)(init_img, resolution=resolution, detect_resolution=resolution)
                    control = result["control"]
                elif control_type == "seg":
                    result = Seg(init_as_control)(init_img, resolution=resolution, detect_resolution=resolution)
                    control = result["control"]
                elif control_type == "normal":
                    result = Normal(init_as_control)(init_img, resolution=resolution)
                    control = result["control"]
                else:
                    raise ValueError(f"Invalid control_type: {control_type}")
                control: torch.Tensor = control.to(self.model["device"])
                H, W, C = result["shape"]
            elif init_img is not None:
                init_img = resize_image(resize_mode, init_img, width, height)
                hires_fix = False
            else:
                if hires_fix and width > 512 and height > 512:
                    logger.debug("HiRes Fix Requested")
                    final_width = width
                    final_height = height
                    if self.model_baseline == "stable diffusion 2":
                        first_pass_ratio = min(final_height / 768, final_width / 768)
                    else:
                        first_pass_ratio = min(final_height / 512, final_width / 512)
                    width = (int(final_width / first_pass_ratio) // 64) * 64
                    height = (int(final_height / first_pass_ratio) // 64) * 64
                    logger.debug(f"First pass image will be processed at width={width}; height={height}")
                else:
                    hires_fix = False
            if mask_mode == "mask":
                if init_mask:
                    init_mask = process_init_mask(init_mask)
            elif mask_mode == "invert":
                if init_mask:
                    init_mask = process_init_mask(init_mask)
                    init_mask = ImageOps.invert(init_mask)
            elif mask_mode == "alpha":
                init_img_transparency = init_img.split()[-1].convert(
                    "L"
                )  # .point(lambda x: 255 if x > 0 else 0, mode='1')
                init_mask = init_img_transparency
                init_mask = init_mask.convert("RGB")
                init_mask = resize_image(resize_mode, init_mask, width, height)
                init_mask = init_mask.convert("RGB")

            """
            vram
            control net doesn't need first stage model on gpu
            regular does, for img2img and hires_fix, if txt2img keep it on cpu
            """
            if control_type is not None:
                low_vram(
                    [
                        (self.control_net_model, "cpu"),
                        (self.control_net_model.control_model, "cpu"),
                        (
                            model.cond_stage_model.transformer
                            if hasattr(model.cond_stage_model, "transformer")
                            else model.cond_stage_model.model.transformer,
                            self.model["device"],
                        ),
                        (self.control_net_model.first_stage_model, "cpu"),
                    ],
                    force=True,  # always move control net to gpu
                )
                self.control_net_model.cond_stage_model.device = self.model["device"]
            else:
                """
                k-diffusion needs the model to be on gpu to create the model wrap
                so we move it to gpu here
                """
                low_vram(
                    [
                        (model, self.model["device"]),
                        (
                            model.cond_stage_model.transformer
                            if hasattr(model.cond_stage_model, "transformer")
                            else model.cond_stage_model.model.transformer,
                            self.model["device"],
                        ),
                        (model.first_stage_model, self.model["device"]),
                    ],
                )
                if low_vram_mode():
                    model.cond_stage_model.device = self.model["device"]

            assert 0.0 <= denoising_strength <= 1.0, "can only work with strength in [0.0, 1.0]"
            t_enc = int(denoising_strength * ddim_steps)

            if (
                init_mask is not None
                and (noise_mode == "matched" or noise_mode == "find_and_matched")
                and init_img is not None
            ):
                noise_q = 0.99
                color_variation = 0.0
                mask_blend_factor = 1.0

                np_init = (np.asarray(init_img.convert("RGB")) / 255.0).astype(
                    np.float64
                )  # annoyingly complex mask fixing
                np_mask_rgb = 1.0 - (np.asarray(ImageOps.invert(init_mask).convert("RGB")) / 255.0).astype(np.float64)
                np_mask_rgb -= np.min(np_mask_rgb)
                np_mask_rgb /= np.max(np_mask_rgb)
                np_mask_rgb = 1.0 - np_mask_rgb
                np_mask_rgb_hardened = 1.0 - (np_mask_rgb < 0.99).astype(np.float64)
                blurred = skimage.filters.gaussian(np_mask_rgb_hardened[:], sigma=16.0, channel_axis=2, truncate=32.0)
                blurred2 = skimage.filters.gaussian(np_mask_rgb_hardened[:], sigma=16.0, channel_axis=2, truncate=32.0)
                # np_mask_rgb_dilated = np_mask_rgb + blurred  # fixup mask todo: derive magic constants
                # np_mask_rgb = np_mask_rgb + blurred
                np_mask_rgb_dilated = np.clip((np_mask_rgb + blurred2) * 0.7071, 0.0, 1.0)
                np_mask_rgb = np.clip((np_mask_rgb + blurred) * 0.7071, 0.0, 1.0)

                noise_rgb = get_matched_noise(np_init, np_mask_rgb, noise_q, color_variation)
                blend_mask_rgb = np.clip(np_mask_rgb_dilated, 0.0, 1.0) ** (mask_blend_factor)
                noised = noise_rgb[:]
                blend_mask_rgb **= 2.0
                noised = np_init[:] * (1.0 - blend_mask_rgb) + noised * blend_mask_rgb

                np_mask_grey = np.sum(np_mask_rgb, axis=2) / 3.0
                ref_mask = np_mask_grey < 1e-3

                all_mask = np.ones((height, width), dtype=bool)
                noised[all_mask, :] = skimage.exposure.match_histograms(
                    noised[all_mask, :] ** 1.0, noised[ref_mask, :], channel_axis=1
                )

                init_img = Image.fromarray(np.clip(noised * 255.0, 0.0, 255.0).astype(np.uint8), mode="RGB")

            def init(model: Union[LatentDiffusion, LatentDiffusionPix2Pix, ControlLDM], init_img: Image.Image):
                image = init_img.convert("RGB")
                image = np.array(image).astype(np.float32) / 255.0
                image = image[None].transpose(0, 3, 1, 2)
                image = torch.from_numpy(image)

                mask_channel = None
                if init_mask:
                    alpha = resize_image(resize_mode, init_mask, width // 8, height // 8)
                    mask_channel = alpha.split()[-1]

                mask = None
                if mask_channel is not None:
                    mask = np.array(mask_channel).astype(np.float32) / 255.0
                    mask = 1 - mask
                    mask = np.tile(mask, (4, 1, 1))
                    mask = mask[None].transpose(0, 1, 2, 3)
                    mask = torch.from_numpy(mask).to(model.first_stage_model.device)

                init_image = 2.0 * image - 1.0
                init_image = init_image.to(model.first_stage_model.device)
                init_latent = model.get_first_stage_encoding(
                    model.encode_first_stage(init_image)
                )  # move to latent space

                return (
                    init_latent,
                    mask,
                )

            def sample_img2img(
                init_data,
                ddim_steps,
                x,
                conditioning,
                unconditional_conditioning,
                sampler_name,
                batch_size=1,
                shape=None,
                karras=False,
                sigma_override: dict = None,
            ):
                start_sampling = time.time()
                nonlocal sampler
                t_enc_steps = t_enc
                if hires_fix:
                    ddim_steps = 20
                    t_enc_steps = int(0.2 * ddim_steps)
                else:
                    t_enc_steps = t_enc
                obliterate = False
                if ddim_steps == t_enc_steps:
                    t_enc_steps = t_enc_steps - 1
                    obliterate = True

                if sampler_name != "DDIM":
                    samples_ddim, _ = sampler.sample_img2img(
                        init_data=init_data,
                        S=ddim_steps,
                        t_enc=t_enc_steps,
                        obliterate=obliterate,
                        conditioning=conditioning,
                        unconditional_conditioning=unconditional_conditioning,
                        unconditional_guidance_scale=cfg_scale,
                        x_T=x,
                    )
                else:
                    x0, z_mask = init_data

                    sampler.make_schedule(ddim_num_steps=ddim_steps, ddim_eta=0.0, verbose=False)
                    z_enc = sampler.stochastic_encode(
                        x0,
                        torch.tensor([t_enc_steps] * batch_size).to(self.model["device"]),
                    )

                    # Obliterate masked image
                    if z_mask is not None and obliterate:
                        random = torch.randn(z_mask.shape, device=z_enc.device)
                        z_enc = (z_mask * random) + ((1 - z_mask) * z_enc)

                        # decode it
                    samples_ddim = sampler.decode(
                        z_enc,
                        conditioning,
                        t_enc_steps,
                        unconditional_guidance_scale=cfg_scale,
                        unconditional_conditioning=unconditional_conditioning,
                    )
                its = round(ddim_steps / (time.time() - start_sampling), 2)
                if hires_fix:
                    logger.info(
                        f"Processed {self.model_name} HiRes-Fix @ {final_width}x{final_height} for {t_enc_steps} iterations at {its} it/s"
                    )
                else:
                    logger.info(
                        f"Processed {self.model_name} img2img @ {width}x{height} for {ddim_steps} iterations at {its} it/s"
                    )
                return samples_ddim

            def sample(
                init_data,
                x,
                conditioning,
                unconditional_conditioning,
                sampler_name,
                batch_size=1,
                shape=None,
                karras=False,
                sigma_override: dict = None,
            ):
                start_sampling = time.time()
                if sampler_name in ["dpmsolver", "DDIM"]:
                    samples_ddim, _ = sampler.sample(
                        ddim_steps,
                        batch_size,
                        shape,
                        conditioning=conditioning,
                        unconditional_guidance_scale=cfg_scale,
                        unconditional_conditioning=unconditional_conditioning,
                        x_T=x,
                        karras=karras,
                        sigma_override=sigma_override,
                    )
                else:
                    samples_ddim, _ = sampler.sample(
                        S=ddim_steps,
                        conditioning=conditioning,
                        unconditional_guidance_scale=cfg_scale,
                        unconditional_conditioning=unconditional_conditioning,
                        x_T=x,
                        karras=karras,
                        sigma_override=sigma_override,
                    )
                its = round(ddim_steps / (time.time() - start_sampling), 2)
                logger.info(
                    f"Processed {self.model_name} txt2img @ {width}x{height} for {ddim_steps} iterations at {its} it/s"
                )
                return samples_ddim

            def create_sampler_by_sampler_name(model):
                nonlocal sampler_name
                if self.model_baseline == "stable diffusion 2":
                    v = True
                    # model_baseline = "stable diffusion 2" covers sd2.x 768 models
                else:
                    # 1.x models and 2.x 512 models do not use v-prediction
                    v = False
                if (
                    sampler_name == "PLMS" and "stable diffusion 2" not in self.model_baseline
                ):  # TODO: check support for sd2.x
                    sampler = PLMSSampler(model)
                elif sampler_name == "DDIM":
                    sampler = DDIMSampler(model)
                elif sampler_name == "k_dpm_2_a":
                    sampler = KDiffusionSampler(model, "dpm_2_ancestral", v=v)
                elif sampler_name == "k_dpm_2":
                    sampler = KDiffusionSampler(model, "dpm_2", v=v)
                elif sampler_name == "k_euler_a":
                    sampler = KDiffusionSampler(model, "euler_ancestral", v=v)
                elif sampler_name == "k_euler":
                    sampler = KDiffusionSampler(model, "euler", v=v)
                elif sampler_name == "k_heun":
                    sampler = KDiffusionSampler(model, "heun", v=v)
                elif sampler_name == "k_lms":
                    sampler = KDiffusionSampler(model, "lms", v=v)
                elif sampler_name == "k_dpm_fast":
                    sampler = KDiffusionSampler(model, "dpm_fast", v=v)
                elif sampler_name == "k_dpm_adaptive":
                    sampler = KDiffusionSampler(model, "dpm_adaptive", v=v)
                elif sampler_name == "k_dpmpp_2s_a":
                    sampler = KDiffusionSampler(model, "dpmpp_2s_ancestral", v=v)
                elif sampler_name == "k_dpmpp_2m":
                    sampler = KDiffusionSampler(model, "dpmpp_2m", v=v)
                elif sampler_name == "k_dpmpp_sde":
                    sampler = KDiffusionSampler(model, "dpmpp_sde", v=v)
                elif sampler_name == "dpmsolver" and "stable diffusion 2" in self.model_baseline:  # only for sd2.x
                    sampler = DPMSolverSampler(model)
                else:
                    logger.error(f"Sampler '{sampler_name}' unknown or does not match model.")
                    # Ensure we don't hit an UnboundLocalError because sampler is not set
                    raise Exception(f"Sampler name not found of does not match model baseline {self.model_baseline}")
                return sampler

            seed = seed_to_int(seed)

            image_dict = {"seed": seed}
            negprompt = ""
            if "###" in prompt:
                prompt, negprompt = prompt.split("###", 1)
                prompt = prompt.strip()
                negprompt = negprompt.strip()

            os.makedirs(self.output_dir, exist_ok=True)

            sample_path = os.path.join(self.output_dir, "samples")
            os.makedirs(sample_path, exist_ok=True)

            karras = False
            if "karras" in sampler_name:
                karras = True
                sampler_name = sampler_name.replace("_karras", "")

            if control_type is None:
                for m in model.modules():
                    if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
                        m.padding_mode = "circular" if tiling else m._orig_padding_mode
            sampler = create_sampler_by_sampler_name(model if control_type is None else self.control_net_model)
            if self.load_concepts and self.concepts_dir is not None:
                prompt_tokens = re.findall("<([a-zA-Z0-9-]+)>", prompt)
                if prompt_tokens:
                    process_prompt_tokens(
                        prompt_tokens, model if control_type is None else self.control_net_model, self.concepts_dir
                    )

            all_prompts = batch_size * n_iter * [prompt]
            all_seeds = [seed + x for x in range(len(all_prompts))]
            if control_type is None:
                if inpainting:
                    samples_ddim = inpaint(
                        sampler,
                        init_img,
                        init_mask,
                        prompt,
                        seed,
                        cfg_scale,
                        ddim_steps,
                        n_iter,
                        width,
                        height,
                        negprompt,
                        clip_skip,
                    )
                    prompts = all_prompts
                    seeds = all_seeds
                elif self.model_name != "pix2pix":
                    logger.debug(
                        f"model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                    )
                    with torch.no_grad():
                        for n in range(n_iter):
                            logger.debug(f"Iteration: {n+1}/{n_iter}")
                            prompts = all_prompts[n * batch_size : (n + 1) * batch_size]
                            seeds = all_seeds[n * batch_size : (n + 1) * batch_size]
                            uc = get_learned_conditioning_with_prompt_weights(negprompt, model, clip_skip)

                            if isinstance(prompts, tuple):
                                prompts = list(prompts)

                            c = torch.cat(
                                [
                                    get_learned_conditioning_with_prompt_weights(prompt, model, clip_skip)
                                    for prompt in prompts
                                ]
                            )

                            opt_C = 4
                            opt_f = 8
                            shape = [opt_C, height // opt_f, width // opt_f]
                            if noise_mode in ["find", "find_and_matched"]:
                                x = torch.cat(
                                    batch_size
                                    * [
                                        find_noise_for_image(
                                            model,
                                            self.model["device"],
                                            init_img.convert("RGB"),
                                            "",
                                            find_noise_steps,
                                            0.0,
                                            normalize=True,
                                            clip_skip=clip_skip,
                                        )
                                    ],
                                    dim=0,
                                )
                            else:
                                x = create_random_tensors(shape, seeds=seeds, device=self.model["device"])
                            init_data = init(model, init_img) if init_img else None
                            if sampler_name != "DDIM":
                                low_vram(
                                    [
                                        (model, self.model["device"]),
                                        (
                                            model.cond_stage_model.transformer
                                            if hasattr(model.cond_stage_model, "transformer")
                                            else model.cond_stage_model.model.transformer,
                                            "cpu",
                                        ),
                                        (model.first_stage_model, "cpu"),
                                    ],
                                )
                            if low_vram_mode():
                                model.cond_stage_model.device = "cpu"
                            logger.debug(
                                f"[Low VRAM] before sampling - model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                            )
                            samples_ddim = (
                                sample_img2img(
                                    init_data=init_data,
                                    ddim_steps=ddim_steps,
                                    x=x,
                                    conditioning=c,
                                    unconditional_conditioning=uc,
                                    sampler_name=sampler_name,
                                )
                                if init_img
                                else sample(
                                    init_data=init_data,
                                    x=x,
                                    conditioning=c,
                                    unconditional_conditioning=uc,
                                    sampler_name=sampler_name,
                                    karras=karras,
                                    batch_size=batch_size,
                                    shape=shape,
                                    sigma_override=sigma_override,
                                )
                            )
                            low_vram(
                                [
                                    (model, "cpu"),
                                    (model.first_stage_model, self.model["device"]),
                                ],
                            )
                            logger.debug(
                                f"[Low VRAM] after sampling - model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                            )
                            if hires_fix:
                                logger.debug(
                                    f"[Low VRAM] hires fix start - model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                                )
                                # Put the image back together
                                temp_x = model.decode_first_stage(samples_ddim)
                                temp_x_samples_ddim = torch.clamp((temp_x + 1.0) / 2.0, min=0.0, max=1.0)
                                for i, x_sample in enumerate(temp_x_samples_ddim):
                                    x_sample = 255.0 * rearrange(x_sample.cpu().numpy(), "c h w -> h w c")
                                    x_sample = x_sample.astype(np.uint8)
                                    temp_image = Image.fromarray(x_sample)

                                # Resize Image to final dimensions
                                temp_image = ImageOps.fit(
                                    temp_image, (final_width, final_height), method=Image.Resampling.LANCZOS
                                )
                                shape = [opt_C, final_height // opt_f, final_width // opt_f]
                                x = create_random_tensors(shape, seeds=seeds, device=self.model["device"])

                                # Re-initialise the image
                                init_data_temp = init(model, temp_image)

                                # Send image for img2img processing
                                logger.debug("Hi-Res Fix Pass")
                                low_vram(
                                    [
                                        (model, self.model["device"]),
                                        (model.first_stage_model, "cpu"),
                                    ],
                                )
                                logger.debug(
                                    f"[Low VRAM] hi-res fix -  before sampling - model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                                )
                                samples_ddim = sample_img2img(
                                    init_data=init_data_temp,
                                    ddim_steps=ddim_steps,
                                    x=x,
                                    conditioning=c,
                                    unconditional_conditioning=uc,
                                    sampler_name=sampler_name,
                                )
                                low_vram(
                                    [
                                        (model, "cpu"),
                                        (model.first_stage_model, self.model["device"]),
                                    ],
                                )
                                logger.debug(
                                    f"[Low VRAM] hi-res fix -  before sampling - model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                                )
                else:  # self.model_name is pix2pix
                    logger.debug(
                        f"[Low VRAM] Pix2pix start - model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                    )
                    init_image = init_img
                    init_image = ImageOps.fit(init_image, (width, height), method=Image.Resampling.LANCZOS).convert(
                        "RGB"
                    )
                    null_token = model.get_learned_conditioning([""], 1)
                    with torch.no_grad():
                        for n in range(n_iter):
                            logger.debug(f"Iteration: {n+1}/{n_iter}")
                            prompts = all_prompts[n * batch_size : (n + 1) * batch_size]
                            seeds = all_seeds[n * batch_size : (n + 1) * batch_size]

                            cond = {}
                            c = torch.cat(
                                [
                                    get_learned_conditioning_with_prompt_weights(prompt, model, clip_skip)
                                    for prompt in prompts
                                ]
                            )
                            cond["c_crossattn"] = [c]
                            init_image = 2 * torch.tensor(np.array(init_image)).float() / 255 - 1
                            init_image = rearrange(init_image, "h w c -> 1 c h w").to(self.model["device"])
                            cond["c_concat"] = [model.encode_first_stage(init_image).mode()]

                            uncond = {}
                            uncond["c_crossattn"] = [null_token]
                            uncond["c_concat"] = [torch.zeros_like(cond["c_concat"][0])]

                            init_data = init(model, init_img) if init_img else None
                            x0, z_mask = init_data

                            extra_args = {
                                "cond": cond,
                                "uncond": uncond,
                                "text_cfg_scale": cfg_scale,
                                "image_cfg_scale": denoising_strength * 2,
                                "mask": z_mask,
                                "x0": x0,
                            }

                            torch.manual_seed(seed)
                            z = torch.randn_like(cond["c_concat"][0])
                            low_vram(
                                [
                                    (model, self.model["device"]),
                                    (
                                        model.cond_stage_model.transformer
                                        if hasattr(model.cond_stage_model, "transformer")
                                        else model.cond_stage_model.model.transformer,
                                        "cpu",
                                    ),
                                    (model.first_stage_model, "cpu"),
                                ],
                            )
                            logger.debug(
                                f"[Low VRAM] pix2pix - before sampling - model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                            )
                            if sampler_name == "DDIM":
                                samples_ddim, _ = sampler.sample(
                                    S=ddim_steps,
                                    batch_size=batch_size,
                                    shape=shape,
                                    conditioning=extra_args["cond"],
                                    unconditional_guidance_scale=extra_args["text_cfg_scale"],
                                    unconditional_conditioning=extra_args["uncond"],
                                    x_T=z,
                                    karras=karras,
                                    sigma_override=sigma_override,
                                    extra_args=extra_args,
                                )
                            else:
                                samples_ddim, _ = sampler.sample(
                                    S=ddim_steps,
                                    conditioning=extra_args["cond"],
                                    unconditional_guidance_scale=extra_args["text_cfg_scale"],
                                    unconditional_conditioning=extra_args["uncond"],
                                    x_T=z,
                                    karras=karras,
                                    sigma_override=sigma_override,
                                    extra_args=extra_args,
                                )
                            low_vram(
                                [
                                    (model, "cpu"),
                                    (model.first_stage_model, self.model["device"]),
                                ],
                            )
                            logger.debug(
                                f"[Low VRAM] pix2pix - before sampling - model.device = {model.device}, model.cond_stage_model.device = {model.cond_stage_model.device}, model.first_stage_model.device = {model.first_stage_model.device}"
                            )
            else:
                if not return_control_map:
                    if self.model_baseline == "stable diffusion 1":
                        logger.debug(
                            f"[Low VRAM] controlnet start - control_net_model.device = {self.control_net_model.device}, model.cond_stage_model.device = {model.cond_stage_model.device},  model.cond_stage_model.transformer.device = {self.control_net_model.cond_stage_model.transformer.device}, model.first_stage_model.device = {self.control_net_model.first_stage_model.device}"
                        )
                    else:
                        logger.debug("[Low VRAM] controlnet start")
                    with torch.no_grad():
                        for n in range(n_iter):
                            prompts = all_prompts[n * batch_size : (n + 1) * batch_size]
                            seeds = all_seeds[n * batch_size : (n + 1) * batch_size]
                            logger.debug(f"Iteration: {n+1}/{n_iter}")
                            # Force all components to gpu before doing prompt weighting
                            low_vram(
                                [
                                    (self.control_net_model, self.model["device"]),
                                    (self.control_net_model.control_model, self.model["device"]),
                                    (
                                        model.cond_stage_model.transformer
                                        if hasattr(model.cond_stage_model, "transformer")
                                        else model.cond_stage_model.model.transformer,
                                        self.model["device"],
                                    ),
                                    (self.control_net_model.first_stage_model, self.model["device"]),
                                ],
                                force=True,
                            )
                            """
                            NOTE:
                            Use `self.control_net_model` instead of `model` for the control net
                            """
                            cond = {
                                "c_concat": [control],
                                "c_crossattn": [
                                    get_learned_conditioning_with_prompt_weights(
                                        prompt, self.control_net_model, clip_skip
                                    )
                                ],
                            }
                            un_cond = {
                                "c_concat": [control],
                                "c_crossattn": [
                                    get_learned_conditioning_with_prompt_weights(
                                        negprompt, self.control_net_model, clip_skip
                                    )
                                ],
                            }

                            if cond["c_crossattn"][0].shape[1] != un_cond["c_crossattn"][0].shape[1]:
                                cond["c_crossattn"][0], un_cond["c_crossattn"][0] = fix_mismatched_tensors(
                                    cond["c_crossattn"][0], un_cond["c_crossattn"][0], self.control_net_model
                                )
                            shape = (4, H // 8, W // 8)
                            logger.info(f"shape from ControlNet = {shape}")
                            logger.info(f"shape from Source Image = {(4, height // 8, width // 8)}")
                            self.control_net_model.control_scales = [denoising_strength * 2] * 13
                            low_vram(
                                [
                                    (self.control_net_model, self.model["device"]),
                                    (self.control_net_model.control_model, self.model["device"]),
                                    (
                                        model.cond_stage_model.transformer
                                        if hasattr(model.cond_stage_model, "transformer")
                                        else model.cond_stage_model.model.transformer,
                                        self.model["device"],
                                    ),
                                    (self.control_net_model.first_stage_model, "cpu"),
                                ],
                                force=True,
                            )
                            samples_ddim, _ = sampler.sample(
                                ddim_steps,
                                n_iter,
                                shape,
                                cond,
                                verbose=False,
                                eta=ddim_eta,
                                unconditional_guidance_scale=cfg_scale,
                                unconditional_conditioning=un_cond,
                            )
                            low_vram(
                                [
                                    (self.control_net_model, "cpu"),
                                    (self.control_net_model.control_model, "cpu"),
                                    (self.control_net_model.first_stage_model, self.model["device"]),
                                ],
                                force=True,
                            )

            logger.debug(
                f"[Low VRAM] decode - model.first_stage_model.device = {model.first_stage_model.device if control_type is None else self.control_net_model.first_stage_model.device}"
            )

            if control_type is None:
                x_samples_ddim = model.decode_first_stage(samples_ddim)
                x_samples_ddim = torch.clamp((x_samples_ddim + 1.0) / 2.0, min=0.0, max=1.0)
            else:
                if not return_control_map:
                    x_samples_ddim = self.control_net_model.decode_first_stage(samples_ddim)
                    x_samples_ddim = (
                        (einops.rearrange(x_samples_ddim, "b c h w -> b h w c") * 127.5 + 127.5)
                        .cpu()
                        .numpy()
                        .clip(0, 255)
                        .astype(np.uint8)
                    )
                else:
                    x_samples_ddim = result["detected_map"]

        for i, x_sample in enumerate(x_samples_ddim):
            sanitized_prompt = slugify(prompt)
            full_path = os.path.join(os.getcwd(), sample_path)
            sample_path_i = sample_path
            base_count = get_next_sequence_number(sample_path_i)
            if karras:
                sampler_name += "_karras"
            if not return_control_map:
                filename = f"{base_count:05}-{ddim_steps}_{sampler_name}_{seeds[i]}_{sanitized_prompt}"[
                    : 200 - len(full_path)
                ]
            else:
                filename = f"{base_count:05}-{control_type}_detected_map"[: 200 - len(full_path)]
            if control_type is None:
                x_sample = 255.0 * rearrange(x_sample.cpu().numpy(), "c h w -> h w c")
                x_sample = x_sample.astype(np.uint8)
                image = Image.fromarray(x_sample)
            else:
                image = Image.fromarray(x_sample)
            if self.safety_checker is not None and self.filter_nsfw:
                image_features = self.feature_extractor(image, return_tensors="pt").to(self.safety_checker.device)
                output_images, has_nsfw_concept = self.safety_checker(
                    clip_input=image_features.pixel_values, images=x_sample
                )
                if has_nsfw_concept and True in has_nsfw_concept:
                    logger.info(f"Image {filename} has NSFW concept")
                    image = Image.new("RGB", (512, 512))
                    image_dict["censored"] = True
            image_dict["image"] = image
            self.images.append(image_dict)

            if save_individual_images:
                path = os.path.join(sample_path, filename + "." + self.save_extension)
                success = save_sample(image, filename, sample_path_i, self.save_extension)
                if success:
                    if self.output_file_path:
                        self.output_images.append(path)
                    else:
                        self.output_images.append(image)
                else:
                    return

        del sampler
        if control_type is not None:
            self.control_net_manager.unload_model(loaded_control_ldm)
            del self.control_net_model
            low_vram(
                [
                    (model, self.model["device"]),
                    (
                        model.cond_stage_model.transformer
                        if hasattr(model.cond_stage_model, "transformer")
                        else model.cond_stage_model.model.transformer,
                        self.model["device"],
                    ),
                    (model.first_stage_model, self.model["device"]),
                ],
                force=True,
            )
            model.cond_stage_model.device = self.model["device"]
            if not self.disable_voodoo:
                del model  # cleanup voodoo model
        torch.cuda.empty_cache()
        return


def make_batch_sd(image, mask, txt, device, num_samples=1):
    image = np.array(image.convert("RGB"))
    image = image[None].transpose(0, 3, 1, 2)
    image = torch.from_numpy(image).to(dtype=torch.float16) / 127.5 - 1.0

    mask = np.array(mask.convert("L"))
    mask = mask.astype(np.float16) / 255.0
    mask = mask[None, None]
    mask[mask < 0.5] = 0
    mask[mask >= 0.5] = 1
    mask = torch.from_numpy(mask)

    masked_image = image * (mask < 0.5)

    batch = {
        "image": einops.repeat(image.to(device=device), "1 ... -> n ...", n=num_samples),
        "txt": num_samples * [txt],
        "mask": einops.repeat(mask.to(device=device), "1 ... -> n ...", n=num_samples),
        "masked_image": einops.repeat(masked_image.to(device=device), "1 ... -> n ...", n=num_samples),
    }
    return batch


def inpaint(
    sampler, image, mask, prompt, seed, scale, ddim_steps, num_samples, W, H, negative_prompt, clip_skip
) -> Image:
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    model = sampler.model

    prng = np.random.RandomState(seed)
    start_code = prng.randn(num_samples, 4, H // 8, W // 8)
    start_code = torch.from_numpy(start_code).to(device=device, dtype=torch.float32)

    with torch.no_grad():
        with torch.autocast("cuda"):
            batch = make_batch_sd(image, mask, txt=prompt, device=device, num_samples=num_samples)

            c = model.cond_stage_model.encode(batch["txt"], clip_skip)

            c_cat = list()
            for ck in model.concat_keys:
                cc = batch[ck].float()
                if ck != model.masked_image_key:
                    bchw = [num_samples, 4, H // 8, W // 8]
                    cc = torch.nn.functional.interpolate(cc, size=bchw[-2:])
                else:
                    cc = model.get_first_stage_encoding(model.encode_first_stage(cc))
                c_cat.append(cc)
            c_cat = torch.cat(c_cat, dim=1)

            # cond
            cond = {"c_concat": [c_cat], "c_crossattn": [c]}

            # uncond cond
            uc_cross = model.get_unconditional_conditioning(num_samples, negative_prompt, clip_skip)
            uc_full = {"c_concat": [c_cat], "c_crossattn": [uc_cross]}

            shape = [model.channels, H // 8, W // 8]
            samples_ddim, _ = sampler.sample(
                ddim_steps,
                num_samples,
                shape,
                cond,
                verbose=False,
                eta=1.0,
                unconditional_guidance_scale=scale,
                unconditional_conditioning=uc_full,
                x_T=start_code,
            )
            return samples_ddim

            # x_samples_ddim = model.decode_first_stage(samples_cfg)

            # result = torch.clamp((x_samples_ddim+1.0)/2.0,
            #                      min=0.0, max=1.0)

            # result = result.cpu().numpy().transpose(0,2,3,1)

            # img = Image.fromarray((result[0]*255).astype(np.uint8))
            # return img



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import hashlib
import json
import os
import sqlite3
import sys
from pathlib import Path

from PIL import Image
from tqdm import tqdm

from nataili import disable_progress
from nataili.util.logger import logger

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


def get_package():
    pkg = importlib_resources.files("annotator")
    return pkg


def get_cache_directory():
    """The nataili specific directory for caching."""
    NATAILI_CACHE_HOME = os.environ.get("NATAILI_CACHE_HOME")
    base_dir = ""
    if NATAILI_CACHE_HOME:
        base_dir = NATAILI_CACHE_HOME
    else:
        base_dir = os.environ.get("XDG_CACHE_HOME", os.path.join(Path.home(), ".cache/"))
    return os.path.join(base_dir, "nataili")


class Cache:
    def __init__(self, cache_name, cache_subname=None, cache_parentname=None):
        """
        :param cache_name: Name of the cache
        :param cache_subname: Subfolder in the cache
        :param cache_parentname: Parent folder of the cache
        Examples:
        cache = Cache("test", "sub", "parent")
        path = self.path + "/parent/test/sub"

        cache = Cache("test", "sub")
        path = self.path + "/test/sub"

        cache = Cache("test")
        path = self.path + "/test"

        cache = Cache("test", cache_parentname="parent")
        path = self.path + "/parent/test"

        If cache file does not exist it is created
        If cache folder does not exist it is created
        """
        self.path = get_cache_directory()
        if cache_parentname:
            self.path = os.path.join(self.path, cache_parentname)
        self.cache_dir = os.path.join(self.path, cache_name)
        if cache_subname:
            self.cache_dir = os.path.join(self.cache_dir, cache_subname)
        self.cache_db = os.path.join(self.cache_dir, "cache.db")
        logger.debug(f"Cache file: {self.cache_db}")
        logger.debug(f"Cache dir: {self.cache_dir}")
        os.makedirs(self.cache_dir, exist_ok=True)
        self.conn = sqlite3.connect(self.cache_db)
        self.cursor = self.conn.cursor()
        self.create_sqlite_db()

    def list_dir(self, input_directory, extensions=[".webp"]):
        """
        List all files in a directory
        :param input_directory: Directory to list
        :param extensions: List of extensions to filter for
        :return: List of files
        """
        files = []
        for file in tqdm(os.listdir(input_directory), disable=disable_progress.active):
            if os.path.splitext(file)[1] in extensions:
                files.append(os.path.splitext(file)[0])
        return files

    def get_all(self):
        """
        Get all entries from the cache
        :return: List of all entries
        """
        self.cursor.execute("SELECT file FROM cache")
        return [x[0] for x in self.cursor.fetchall()]

    def get_all_export(self):
        self.cursor.execute("SELECT file, pil_hash FROM cache")
        return {x[0]: x[1] for x in self.cursor.fetchall()}

    def filter_list(self, input_list):
        """
        Filter a list
        :param input_list: List to filter
        :param filter_list: List to filter with
        :return: Filtered list
        """
        db_list = self.get_all()
        logger.info(f"Filtering {len(input_list)} files")
        logger.info(f"Filtering {len(db_list)} files")
        logger.info(f"Filtering {len(set(input_list) - set(db_list))} files")
        logger.info(f"First item in input_list: {input_list[0]}")
        logger.info(f"First item in db_list: {db_list[0]}")
        return list(set(input_list) - set(db_list))

    def hash_file(self, file_path):
        """
        Hash a file
        :param file_path: Path to the file
        :return: Hash of the file
        """
        with open(file_path, "rb") as f:
            return hashlib.sha256(f.read()).hexdigest()

    def hash_pil_image(self, pil_image: Image.Image):
        """
        Hash a PIL image
        :param pil_image: PIL image
        :return: Hash of the PIL image
        """
        return hashlib.sha256(pil_image.tobytes()).hexdigest()

    def hash_pil_image_file(self, file_path):
        """
        Hash a PIL image
        :param file_path: Path to the file
        :return: Hash of the PIL image
        """
        pil_image = Image.open(file_path)
        return self.hash_pil_image(pil_image)

    def hash_files(self, files_list, input_directory, extensions=[".webp"]):
        """
        Hash all files in a directory
        :param input_directory: Directory to hash
        :return: List of hashes
        """
        pil_hashes = []
        file_hashes = []
        for file in tqdm(files_list, disable=disable_progress.active):
            for extension in extensions:
                file = file + extension
                file_path = os.path.join(input_directory, file)
                file_hash = self.hash_file(file_path)
                pil_image_hash = self.hash_pil_image_file(file_path)
                pil_hashes.append(pil_image_hash)
                file_hashes.append(file_hash)
        return pil_hashes, file_hashes

    def create_sqlite_db(self):
        """
        Create a sqlite database from the cache
        """
        self.cursor.execute("CREATE TABLE IF NOT EXISTS cache (file text, hash text, pil_hash text)")
        self.cursor.execute("CREATE INDEX IF NOT EXISTS file_index ON cache (file)")
        self.cursor.execute("CREATE INDEX IF NOT EXISTS hash_index ON cache (hash)")
        self.cursor.execute("CREATE INDEX IF NOT EXISTS pil_hash_index ON cache (pil_hash)")
        self.conn.commit()

    def add_sqlite_row(self, file: str, hash: str, pil_hash: str, commit=True):
        """
        Add a row to the sqlite database
        """
        self.cursor.execute("INSERT INTO cache VALUES (?, ?, ?)", (file, hash, pil_hash))
        if commit:
            self.conn.commit()

    def populate_sqlite_db(self, list_of_files: list):
        """
        Populate the sqlite database from the cache
        """
        # Populate sqlite database
        for file in list_of_files:
            self.add_sqlite_row(file["file"], file["hash"], file["pil_hash"], commit=False)
        self.conn.commit()

    def key_exists(self, key):
        """
        Check if a key exists in the cache
        """
        query = "SELECT hash, pil_hash FROM cache WHERE file=?"
        self.cursor.execute(query, (key,))
        if self.cursor.fetchone():
            return True
        return False

    def get(self, file: str = None, file_hash: str = None, pil_hash: str = None, no_return=False):
        """
        Get a file from the cache
        """
        if not any([file, file_hash, pil_hash]):
            raise ValueError("At least one value must be provided to search the database")
        file = os.path.splitext(file)[0] if file else None
        query = "SELECT hash, pil_hash FROM cache WHERE "
        conditions = []
        values = []
        if file:
            conditions.append("file=?")
            values.append(file)
        if file_hash:
            conditions.append("hash=?")
            values.append(file_hash)
        if pil_hash:
            conditions.append("pil_hash=?")
            values.append(pil_hash)

        query += " OR ".join(conditions)

        self.cursor.execute(query, tuple(values))
        result = self.cursor.fetchone()
        if result:
            if no_return:
                return True
            file_hash, pil_hash = result
            if file_hash:
                file_path = os.path.join(self.cache_dir, file_hash + ".npy")
                if os.path.exists(file_path):
                    return file_path
            if pil_hash:
                file_path = os.path.join(self.cache_dir, pil_hash + ".npy")
                if os.path.exists(file_path):
                    return file_path
        return None



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import safetensors.torch
import torch
from omegaconf import OmegaConf

from ldm.util import instantiate_from_config
from nataili.cache import get_cache_directory
from nataili.model_manager.base import BaseModelManager
from nataili.model_manager.compvis import CompVisModelManager, DisableInitialization
from nataili.util.logger import logger


class ControlNetModelManager(BaseModelManager):
    def __init__(self, download_reference=True, compvis=None):
        super().__init__()
        self.download_reference = download_reference
        self.path = f"{get_cache_directory()}/controlnet"
        self.models_db_name = "controlnet"
        self.models_path = self.pkg / f"{self.models_db_name}.json"
        self.remote_db = (
            f"https://raw.githubusercontent.com/db0/AI-Horde-image-model-reference/main/{self.models_db_name}.json"
        )
        self.control_nets = {}
        self.init()

    def load_control_ldm(
        self,
        model_name,
        target_name,
        input_state_dict,
        half_precision=True,
        gpu_id=0,
        cpu_only=False,
        _device=None,
    ):
        if not self.cuda_available:
            cpu_only = True
        if cpu_only:
            device = torch.device("cpu")
            half_precision = False
        else:
            device = torch.device(f"cuda:{gpu_id}" if self.cuda_available else "cpu")
        if model_name not in self.control_nets:
            logger.error(f"{model_name} not loaded")
            return False
        config_path = self.get_model_files(model_name)[1]["path"]
        config_path = f"{self.pkg}/{config_path}"
        logger.info(f"Loading controlLDM {model_name} for {target_name}")
        config = OmegaConf.load(config_path)
        try:
            with DisableInitialization(disable_clip=True):
                model = instantiate_from_config(config.model)
        except Exception as e:
            pass
        full_name = f"{model_name}_{target_name}"
        logger.info(f"Loaded {full_name} ControlLDM")
        sd15_with_control_state_dict = self.control_nets[model_name]["state_dict"]
        final_state_dict: dict = input_state_dict.copy()
        keys = sd15_with_control_state_dict.keys()
        logger.info("Merge control net state dict into target state dict")
        if "_sd2" not in model_name:
            for key in keys:
                if not key.startswith("control_"):
                    continue
                p = sd15_with_control_state_dict[key]
                key_name = f'model.diffusion_model.{key.replace("control_model.", "")}'
                if key_name in input_state_dict.keys():
                    # logger.info(f"merging {key_name} from input {key} from control")
                    p_new = p + input_state_dict[key_name].clone().cpu()
                else:
                    # logger.info(f"directly copying {key_name} from control")
                    p_new = p
                final_state_dict[key] = p_new
        else:
            for key in keys:
                p = sd15_with_control_state_dict[key]
                key_name = f'model.diffusion_model.{key.replace("control_model.", "")}'
                if key in input_state_dict.keys():
                    # logger.info(f"merging {key_name} from input {key} from control")
                    p_new = p + input_state_dict[key_name].clone().cpu()
                else:
                    # logger.info(f"directly copying {key_name} from control")
                    p_new = p
                final_state_dict[f"control_model.{key}"] = p_new
        # remove key "lvlb_weights"
        if "lvlb_weights" in final_state_dict.keys():
            final_state_dict.pop("lvlb_weights")
        logger.info("Finished merging control net state dict into target state dict")
        logger.info(f"Loading {full_name} state dict")
        model.load_state_dict(final_state_dict, strict=True)
        logger.info(f"Loaded {full_name} state dict")
        if half_precision:
            model.half()
        del final_state_dict, sd15_with_control_state_dict, input_state_dict
        self.loaded_models[full_name] = {"model": model, "device": device, "half_precision": half_precision}

    def load_controlnet(
        self,
        model_name,
    ):
        if model_name not in self.models:
            logger.error(f"{model_name} not found")
            return False
        if model_name not in self.available_models:
            logger.error(f"{model_name} not available")
            logger.init_ok(f"Downloading {model_name}", status="Downloading")
            self.download_model(model_name)
            logger.init_ok(f"{model_name} downloaded", status="Downloading")
        if model_name in self.control_nets:
            logger.info(f"{model_name} already loaded")
            return True
        model_path = self.get_model_files(model_name)[0]["path"]
        model_path = f"{self.path}/{model_path}"
        logger.info(f"Loading controlnet {model_name}")
        logger.info(f"Model path: {model_path}")
        state_dict = safetensors.torch.load_file(model_path)

        self.control_nets[model_name] = {"state_dict": state_dict}
        return True



# pytorch_diffusion + derived encoder decoder
import gc
import math

import numpy as np
import torch
import torch.nn as nn
from einops import rearrange

from ldm.modules.attention import LinearAttention
from ldm.util import instantiate_from_config
from nataili.util.logger import logger


def get_timestep_embedding(timesteps, embedding_dim):
    """
    This matches the implementation in Denoising Diffusion Probabilistic Models:
    From Fairseq.
    Build sinusoidal embeddings.
    This matches the implementation in tensor2tensor, but differs slightly
    from the description in Section 3.5 of "Attention Is All You Need".
    """
    assert len(timesteps.shape) == 1

    half_dim = embedding_dim // 2
    emb = math.log(10000) / (half_dim - 1)
    emb = torch.exp(torch.arange(half_dim, dtype=torch.float32) * -emb)
    emb = emb.to(device=timesteps.device)
    emb = timesteps.float()[:, None] * emb[None, :]
    emb = torch.cat([torch.sin(emb), torch.cos(emb)], dim=1)
    if embedding_dim % 2 == 1:  # zero pad
        emb = torch.nn.functional.pad(emb, (0,1,0,0))
    return emb


def nonlinearity(x):
    # swish
    return x*torch.sigmoid(x)


def Normalize(in_channels, num_groups=32):
    return torch.nn.GroupNorm(num_groups=num_groups, num_channels=in_channels, eps=1e-6, affine=True)


class Upsample(nn.Module):
    def __init__(self, in_channels, with_conv):
        super().__init__()
        self.with_conv = with_conv
        if self.with_conv:
            self.conv = torch.nn.Conv2d(in_channels,
                                        in_channels,
                                        kernel_size=3,
                                        stride=1,
                                        padding=1)

    def forward(self, x):
        x = torch.nn.functional.interpolate(x, scale_factor=2.0, mode="nearest")
        if self.with_conv:
            x = self.conv(x)
        return x


class Downsample(nn.Module):
    def __init__(self, in_channels, with_conv):
        super().__init__()
        self.with_conv = with_conv
        if self.with_conv:
            # no asymmetric padding in torch conv, must do it ourselves
            self.conv = torch.nn.Conv2d(in_channels,
                                        in_channels,
                                        kernel_size=3,
                                        stride=2,
                                        padding=0)

    def forward(self, x):
        if self.with_conv:
            pad = (0,1,0,1)
            x = torch.nn.functional.pad(x, pad, mode="constant", value=0)
            x = self.conv(x)
        else:
            x = torch.nn.functional.avg_pool2d(x, kernel_size=2, stride=2)
        return x


class ResnetBlock(nn.Module):
    def __init__(self, *, in_channels, out_channels=None, conv_shortcut=False,
                 dropout, temb_channels=512):
        super().__init__()
        self.in_channels = in_channels
        out_channels = in_channels if out_channels is None else out_channels
        self.out_channels = out_channels
        self.use_conv_shortcut = conv_shortcut

        self.norm1 = Normalize(in_channels)
        self.conv1 = torch.nn.Conv2d(in_channels,
                                     out_channels,
                                     kernel_size=3,
                                     stride=1,
                                     padding=1)
        if temb_channels > 0:
            self.temb_proj = torch.nn.Linear(temb_channels,
                                             out_channels)
        self.norm2 = Normalize(out_channels)
        self.dropout = torch.nn.Dropout(dropout)
        self.conv2 = torch.nn.Conv2d(out_channels,
                                     out_channels,
                                     kernel_size=3,
                                     stride=1,
                                     padding=1)
        if self.in_channels != self.out_channels:
            if self.use_conv_shortcut:
                self.conv_shortcut = torch.nn.Conv2d(in_channels,
                                                     out_channels,
                                                     kernel_size=3,
                                                     stride=1,
                                                     padding=1)
            else:
                self.nin_shortcut = torch.nn.Conv2d(in_channels,
                                                    out_channels,
                                                    kernel_size=1,
                                                    stride=1,
                                                    padding=0)

    def forward(self, x, temb):
        h1 = x
        h2 = self.norm1(h1)
        del h1

        h3 = nonlinearity(h2)
        del h2

        h4 = self.conv1(h3)
        del h3

        if temb is not None:
            h4 = h4 + self.temb_proj(nonlinearity(temb))[:,:,None,None]

        h5 = self.norm2(h4)
        del h4

        h6 = nonlinearity(h5)
        del h5

        h7 = self.dropout(h6)
        del h6

        h8 = self.conv2(h7)
        del h7

        if self.in_channels != self.out_channels:
            if self.use_conv_shortcut:
                x = self.conv_shortcut(x)
            else:
                x = self.nin_shortcut(x)

        return x + h8


class LinAttnBlock(LinearAttention):
    """to match AttnBlock usage"""
    def __init__(self, in_channels):
        super().__init__(dim=in_channels, heads=1, dim_head=in_channels)


class AttnBlock(nn.Module):
    def __init__(self, in_channels):
        super().__init__()
        self.in_channels = in_channels

        self.norm = Normalize(in_channels)
        self.q = torch.nn.Conv2d(in_channels,
                                 in_channels,
                                 kernel_size=1,
                                 stride=1,
                                 padding=0)
        self.k = torch.nn.Conv2d(in_channels,
                                 in_channels,
                                 kernel_size=1,
                                 stride=1,
                                 padding=0)
        self.v = torch.nn.Conv2d(in_channels,
                                 in_channels,
                                 kernel_size=1,
                                 stride=1,
                                 padding=0)
        self.proj_out = torch.nn.Conv2d(in_channels,
                                        in_channels,
                                        kernel_size=1,
                                        stride=1,
                                        padding=0)


    def forward(self, x):
        h_ = x
        h_ = self.norm(h_)
        q1 = self.q(h_)
        k1 = self.k(h_)
        v = self.v(h_)

        # compute attention
        b, c, h, w = q1.shape

        q2 = q1.reshape(b, c, h*w)
        del q1

        q = q2.permute(0, 2, 1)   # b,hw,c
        del q2

        k = k1.reshape(b, c, h*w) # b,c,hw
        del k1

        h_ = torch.zeros_like(k, device=q.device)

        stats = torch.cuda.memory_stats(q.device)
        mem_active = stats['active_bytes.all.current']
        mem_reserved = stats['reserved_bytes.all.current']
        mem_free_cuda, _ = torch.cuda.mem_get_info(torch.cuda.current_device())
        mem_free_torch = mem_reserved - mem_active
        mem_free_total = mem_free_cuda + mem_free_torch

        tensor_size = q.shape[0] * q.shape[1] * k.shape[2] * 4
        mem_required = tensor_size * 2.5
        steps = 1

        if mem_required > mem_free_total:
            steps = 2**(math.ceil(math.log(mem_required / mem_free_total, 2)))

        slice_size = q.shape[1] // steps if (q.shape[1] % steps) == 0 else q.shape[1]
        for i in range(0, q.shape[1], slice_size):
            end = i + slice_size

            w1 = torch.bmm(q[:, i:end], k)     # b,hw,hw    w[b,i,j]=sum_c q[b,i,c]k[b,c,j]
            w2 = w1 * (int(c)**(-0.5))
            del w1
            w3 = torch.nn.functional.softmax(w2, dim=2)
            del w2

            # attend to values
            v1 = v.reshape(b, c, h*w)
            w4 = w3.permute(0, 2, 1)   # b,hw,hw (first hw of k, second of q)
            del w3

            h_[:, :, i:end] = torch.bmm(v1, w4)     # b, c,hw (hw of q) h_[b,c,j] = sum_i v[b,c,i] w_[b,i,j]
            del v1, w4

        h2 = h_.reshape(b, c, h, w)
        del h_

        h3 = self.proj_out(h2)
        del h2

        h3 += x

        return h3


def make_attn(in_channels, attn_type="vanilla"):
    assert attn_type in ["vanilla", "linear", "none"], f'attn_type {attn_type} unknown'
    logger.debug(f"making attention of type '{attn_type}' with {in_channels} in_channels")
    if attn_type == "vanilla":
        return AttnBlock(in_channels)
    elif attn_type == "none":
        return nn.Identity(in_channels)
    else:
        return LinAttnBlock(in_channels)


class Model(nn.Module):
    def __init__(self, *, ch, out_ch, ch_mult=(1,2,4,8), num_res_blocks,
                 attn_resolutions, dropout=0.0, resamp_with_conv=True, in_channels,
                 resolution, use_timestep=True, use_linear_attn=False, attn_type="vanilla"):
        super().__init__()
        if use_linear_attn: attn_type = "linear"
        self.ch = ch
        self.temb_ch = self.ch*4
        self.num_resolutions = len(ch_mult)
        self.num_res_blocks = num_res_blocks
        self.resolution = resolution
        self.in_channels = in_channels

        self.use_timestep = use_timestep
        if self.use_timestep:
            # timestep embedding
            self.temb = nn.Module()
            self.temb.dense = nn.ModuleList([
                torch.nn.Linear(self.ch,
                                self.temb_ch),
                torch.nn.Linear(self.temb_ch,
                                self.temb_ch),
            ])

        # downsampling
        self.conv_in = torch.nn.Conv2d(in_channels,
                                       self.ch,
                                       kernel_size=3,
                                       stride=1,
                                       padding=1)

        curr_res = resolution
        in_ch_mult = (1,)+tuple(ch_mult)
        self.down = nn.ModuleList()
        for i_level in range(self.num_resolutions):
            block = nn.ModuleList()
            attn = nn.ModuleList()
            block_in = ch*in_ch_mult[i_level]
            block_out = ch*ch_mult[i_level]
            for i_block in range(self.num_res_blocks):
                block.append(ResnetBlock(in_channels=block_in,
                                         out_channels=block_out,
                                         temb_channels=self.temb_ch,
                                         dropout=dropout))
                block_in = block_out
                if curr_res in attn_resolutions:
                    attn.append(make_attn(block_in, attn_type=attn_type))
            down = nn.Module()
            down.block = block
            down.attn = attn
            if i_level != self.num_resolutions-1:
                down.downsample = Downsample(block_in, resamp_with_conv)
                curr_res = curr_res // 2
            self.down.append(down)

        # middle
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(in_channels=block_in,
                                       out_channels=block_in,
                                       temb_channels=self.temb_ch,
                                       dropout=dropout)
        self.mid.attn_1 = make_attn(block_in, attn_type=attn_type)
        self.mid.block_2 = ResnetBlock(in_channels=block_in,
                                       out_channels=block_in,
                                       temb_channels=self.temb_ch,
                                       dropout=dropout)

        # upsampling
        self.up = nn.ModuleList()
        for i_level in reversed(range(self.num_resolutions)):
            block = nn.ModuleList()
            attn = nn.ModuleList()
            block_out = ch*ch_mult[i_level]
            skip_in = ch*ch_mult[i_level]
            for i_block in range(self.num_res_blocks+1):
                if i_block == self.num_res_blocks:
                    skip_in = ch*in_ch_mult[i_level]
                block.append(ResnetBlock(in_channels=block_in+skip_in,
                                         out_channels=block_out,
                                         temb_channels=self.temb_ch,
                                         dropout=dropout))
                block_in = block_out
                if curr_res in attn_resolutions:
                    attn.append(make_attn(block_in, attn_type=attn_type))
            up = nn.Module()
            up.block = block
            up.attn = attn
            if i_level != 0:
                up.upsample = Upsample(block_in, resamp_with_conv)
                curr_res = curr_res * 2
            self.up.insert(0, up) # prepend to get consistent order

        # end
        self.norm_out = Normalize(block_in)
        self.conv_out = torch.nn.Conv2d(block_in,
                                        out_ch,
                                        kernel_size=3,
                                        stride=1,
                                        padding=1)

    def forward(self, x, t=None, context=None):
        #assert x.shape[2] == x.shape[3] == self.resolution
        if context is not None:
            # assume aligned context, cat along channel axis
            x = torch.cat((x, context), dim=1)
        if self.use_timestep:
            # timestep embedding
            assert t is not None
            temb = get_timestep_embedding(t, self.ch)
            temb = self.temb.dense[0](temb)
            temb = nonlinearity(temb)
            temb = self.temb.dense[1](temb)
        else:
            temb = None

        # downsampling
        hs = [self.conv_in(x)]
        for i_level in range(self.num_resolutions):
            for i_block in range(self.num_res_blocks):
                h = self.down[i_level].block[i_block](hs[-1], temb)
                if len(self.down[i_level].attn) > 0:
                    h = self.down[i_level].attn[i_block](h)
                hs.append(h)
            if i_level != self.num_resolutions-1:
                hs.append(self.down[i_level].downsample(hs[-1]))

        # middle
        h = hs[-1]
        h = self.mid.block_1(h, temb)
        h = self.mid.attn_1(h)
        h = self.mid.block_2(h, temb)

        # upsampling
        for i_level in reversed(range(self.num_resolutions)):
            for i_block in range(self.num_res_blocks+1):
                h = self.up[i_level].block[i_block](
                    torch.cat([h, hs.pop()], dim=1), temb)
                if len(self.up[i_level].attn) > 0:
                    h = self.up[i_level].attn[i_block](h)
            if i_level != 0:
                h = self.up[i_level].upsample(h)

        # end
        h = self.norm_out(h)
        h = nonlinearity(h)
        h = self.conv_out(h)
        return h

    def get_last_layer(self):
        return self.conv_out.weight


class Encoder(nn.Module):
    def __init__(self, *, ch, out_ch, ch_mult=(1,2,4,8), num_res_blocks,
                 attn_resolutions, dropout=0.0, resamp_with_conv=True, in_channels,
                 resolution, z_channels, double_z=True, use_linear_attn=False, attn_type="vanilla",
                 **ignore_kwargs):
        super().__init__()
        if use_linear_attn: attn_type = "linear"
        self.ch = ch
        self.temb_ch = 0
        self.num_resolutions = len(ch_mult)
        self.num_res_blocks = num_res_blocks
        self.resolution = resolution
        self.in_channels = in_channels

        # downsampling
        self.conv_in = torch.nn.Conv2d(in_channels,
                                       self.ch,
                                       kernel_size=3,
                                       stride=1,
                                       padding=1)

        curr_res = resolution
        in_ch_mult = (1,)+tuple(ch_mult)
        self.in_ch_mult = in_ch_mult
        self.down = nn.ModuleList()
        for i_level in range(self.num_resolutions):
            block = nn.ModuleList()
            attn = nn.ModuleList()
            block_in = ch*in_ch_mult[i_level]
            block_out = ch*ch_mult[i_level]
            for i_block in range(self.num_res_blocks):
                block.append(ResnetBlock(in_channels=block_in,
                                         out_channels=block_out,
                                         temb_channels=self.temb_ch,
                                         dropout=dropout))
                block_in = block_out
                if curr_res in attn_resolutions:
                    attn.append(make_attn(block_in, attn_type=attn_type))
            down = nn.Module()
            down.block = block
            down.attn = attn
            if i_level != self.num_resolutions-1:
                down.downsample = Downsample(block_in, resamp_with_conv)
                curr_res = curr_res // 2
            self.down.append(down)

        # middle
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(in_channels=block_in,
                                       out_channels=block_in,
                                       temb_channels=self.temb_ch,
                                       dropout=dropout)
        self.mid.attn_1 = make_attn(block_in, attn_type=attn_type)
        self.mid.block_2 = ResnetBlock(in_channels=block_in,
                                       out_channels=block_in,
                                       temb_channels=self.temb_ch,
                                       dropout=dropout)

        # end
        self.norm_out = Normalize(block_in)
        self.conv_out = torch.nn.Conv2d(block_in,
                                        2*z_channels if double_z else z_channels,
                                        kernel_size=3,
                                        stride=1,
                                        padding=1)

    def forward(self, x):
        # timestep embedding
        temb = None

        # downsampling
        hs = [self.conv_in(x)]
        for i_level in range(self.num_resolutions):
            for i_block in range(self.num_res_blocks):
                h = self.down[i_level].block[i_block](hs[-1], temb)
                if len(self.down[i_level].attn) > 0:
                    h = self.down[i_level].attn[i_block](h)
                hs.append(h)
            if i_level != self.num_resolutions-1:
                hs.append(self.down[i_level].downsample(hs[-1]))

        # middle
        h = hs[-1]
        h = self.mid.block_1(h, temb)
        h = self.mid.attn_1(h)
        h = self.mid.block_2(h, temb)

        # end
        h = self.norm_out(h)
        h = nonlinearity(h)
        h = self.conv_out(h)
        return h


class Decoder(nn.Module):
    def __init__(self, *, ch, out_ch, ch_mult=(1,2,4,8), num_res_blocks,
                 attn_resolutions, dropout=0.0, resamp_with_conv=True, in_channels,
                 resolution, z_channels, give_pre_end=False, tanh_out=False, use_linear_attn=False,
                 attn_type="vanilla", **ignorekwargs):
        super().__init__()
        if use_linear_attn: attn_type = "linear"
        self.ch = ch
        self.temb_ch = 0
        self.num_resolutions = len(ch_mult)
        self.num_res_blocks = num_res_blocks
        self.resolution = resolution
        self.in_channels = in_channels
        self.give_pre_end = give_pre_end
        self.tanh_out = tanh_out

        # compute in_ch_mult, block_in and curr_res at lowest res
        in_ch_mult = (1,)+tuple(ch_mult)
        block_in = ch*ch_mult[self.num_resolutions-1]
        curr_res = resolution // 2**(self.num_resolutions-1)
        self.z_shape = (1,z_channels,curr_res,curr_res)
        logger.debug("Working with z of shape {} = {} dimensions.".format(
            self.z_shape, np.prod(self.z_shape)))

        # z to block_in
        self.conv_in = torch.nn.Conv2d(z_channels,
                                       block_in,
                                       kernel_size=3,
                                       stride=1,
                                       padding=1)

        # middle
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(in_channels=block_in,
                                       out_channels=block_in,
                                       temb_channels=self.temb_ch,
                                       dropout=dropout)
        self.mid.attn_1 = make_attn(block_in, attn_type=attn_type)
        self.mid.block_2 = ResnetBlock(in_channels=block_in,
                                       out_channels=block_in,
                                       temb_channels=self.temb_ch,
                                       dropout=dropout)

        # upsampling
        self.up = nn.ModuleList()
        for i_level in reversed(range(self.num_resolutions)):
            block = nn.ModuleList()
            attn = nn.ModuleList()
            block_out = ch*ch_mult[i_level]
            for i_block in range(self.num_res_blocks+1):
                block.append(ResnetBlock(in_channels=block_in,
                                         out_channels=block_out,
                                         temb_channels=self.temb_ch,
                                         dropout=dropout))
                block_in = block_out
                if curr_res in attn_resolutions:
                    attn.append(make_attn(block_in, attn_type=attn_type))
            up = nn.Module()
            up.block = block
            up.attn = attn
            if i_level != 0:
                up.upsample = Upsample(block_in, resamp_with_conv)
                curr_res = curr_res * 2
            self.up.insert(0, up) # prepend to get consistent order

        # end
        self.norm_out = Normalize(block_in)
        self.conv_out = torch.nn.Conv2d(block_in,
                                        out_ch,
                                        kernel_size=3,
                                        stride=1,
                                        padding=1)

    def forward(self, z):
        #assert z.shape[1:] == self.z_shape[1:]
        self.last_z_shape = z.shape

        # timestep embedding
        temb = None

        # z to block_in
        h1 = self.conv_in(z)

        # middle
        h2 = self.mid.block_1(h1, temb)
        del h1

        h3 = self.mid.attn_1(h2)
        del h2

        h = self.mid.block_2(h3, temb)
        del h3

        # prepare for up sampling
        gc.collect()
        torch.cuda.empty_cache()

        # upsampling
        for i_level in reversed(range(self.num_resolutions)):
            for i_block in range(self.num_res_blocks+1):
                h = self.up[i_level].block[i_block](h, temb)
                if len(self.up[i_level].attn) > 0:
                    t = h
                    h = self.up[i_level].attn[i_block](t)
                    del t

            if i_level != 0:
                t = h
                h = self.up[i_level].upsample(t)
                del t

        # end
        if self.give_pre_end:
            return h

        h1 = self.norm_out(h)
        del h

        h2 = nonlinearity(h1)
        del h1

        h = self.conv_out(h2)
        del h2

        if self.tanh_out:
            t = h
            h = torch.tanh(t)
            del t

        return h


class SimpleDecoder(nn.Module):
    def __init__(self, in_channels, out_channels, *args, **kwargs):
        super().__init__()
        self.model = nn.ModuleList([nn.Conv2d(in_channels, in_channels, 1),
                                     ResnetBlock(in_channels=in_channels,
                                                 out_channels=2 * in_channels,
                                                 temb_channels=0, dropout=0.0),
                                     ResnetBlock(in_channels=2 * in_channels,
                                                out_channels=4 * in_channels,
                                                temb_channels=0, dropout=0.0),
                                     ResnetBlock(in_channels=4 * in_channels,
                                                out_channels=2 * in_channels,
                                                temb_channels=0, dropout=0.0),
                                     nn.Conv2d(2*in_channels, in_channels, 1),
                                     Upsample(in_channels, with_conv=True)])
        # end
        self.norm_out = Normalize(in_channels)
        self.conv_out = torch.nn.Conv2d(in_channels,
                                        out_channels,
                                        kernel_size=3,
                                        stride=1,
                                        padding=1)

    def forward(self, x):
        for i, layer in enumerate(self.model):
            if i in [1,2,3]:
                x = layer(x, None)
            else:
                x = layer(x)

        h = self.norm_out(x)
        h = nonlinearity(h)
        x = self.conv_out(h)
        return x


class UpsampleDecoder(nn.Module):
    def __init__(self, in_channels, out_channels, ch, num_res_blocks, resolution,
                 ch_mult=(2,2), dropout=0.0):
        super().__init__()
        # upsampling
        self.temb_ch = 0
        self.num_resolutions = len(ch_mult)
        self.num_res_blocks = num_res_blocks
        block_in = in_channels
        curr_res = resolution // 2 ** (self.num_resolutions - 1)
        self.res_blocks = nn.ModuleList()
        self.upsample_blocks = nn.ModuleList()
        for i_level in range(self.num_resolutions):
            res_block = []
            block_out = ch * ch_mult[i_level]
            for i_block in range(self.num_res_blocks + 1):
                res_block.append(ResnetBlock(in_channels=block_in,
                                         out_channels=block_out,
                                         temb_channels=self.temb_ch,
                                         dropout=dropout))
                block_in = block_out
            self.res_blocks.append(nn.ModuleList(res_block))
            if i_level != self.num_resolutions - 1:
                self.upsample_blocks.append(Upsample(block_in, True))
                curr_res = curr_res * 2

        # end
        self.norm_out = Normalize(block_in)
        self.conv_out = torch.nn.Conv2d(block_in,
                                        out_channels,
                                        kernel_size=3,
                                        stride=1,
                                        padding=1)

    def forward(self, x):
        # upsampling
        h = x
        for k, i_level in enumerate(range(self.num_resolutions)):
            for i_block in range(self.num_res_blocks + 1):
                h = self.res_blocks[i_level][i_block](h, None)
            if i_level != self.num_resolutions - 1:
                h = self.upsample_blocks[k](h)
        h = self.norm_out(h)
        h = nonlinearity(h)
        h = self.conv_out(h)
        return h


class LatentRescaler(nn.Module):
    def __init__(self, factor, in_channels, mid_channels, out_channels, depth=2):
        super().__init__()
        # residual block, interpolate, residual block
        self.factor = factor
        self.conv_in = nn.Conv2d(in_channels,
                                 mid_channels,
                                 kernel_size=3,
                                 stride=1,
                                 padding=1)
        self.res_block1 = nn.ModuleList([ResnetBlock(in_channels=mid_channels,
                                                     out_channels=mid_channels,
                                                     temb_channels=0,
                                                     dropout=0.0) for _ in range(depth)])
        self.attn = AttnBlock(mid_channels)
        self.res_block2 = nn.ModuleList([ResnetBlock(in_channels=mid_channels,
                                                     out_channels=mid_channels,
                                                     temb_channels=0,
                                                     dropout=0.0) for _ in range(depth)])

        self.conv_out = nn.Conv2d(mid_channels,
                                  out_channels,
                                  kernel_size=1,
                                  )

    def forward(self, x):
        x = self.conv_in(x)
        for block in self.res_block1:
            x = block(x, None)
        x = torch.nn.functional.interpolate(x, size=(int(round(x.shape[2]*self.factor)), int(round(x.shape[3]*self.factor))))
        x = self.attn(x)
        for block in self.res_block2:
            x = block(x, None)
        x = self.conv_out(x)
        return x


class MergedRescaleEncoder(nn.Module):
    def __init__(self, in_channels, ch, resolution, out_ch, num_res_blocks,
                 attn_resolutions, dropout=0.0, resamp_with_conv=True,
                 ch_mult=(1,2,4,8), rescale_factor=1.0, rescale_module_depth=1):
        super().__init__()
        intermediate_chn = ch * ch_mult[-1]
        self.encoder = Encoder(in_channels=in_channels, num_res_blocks=num_res_blocks, ch=ch, ch_mult=ch_mult,
                               z_channels=intermediate_chn, double_z=False, resolution=resolution,
                               attn_resolutions=attn_resolutions, dropout=dropout, resamp_with_conv=resamp_with_conv,
                               out_ch=None)
        self.rescaler = LatentRescaler(factor=rescale_factor, in_channels=intermediate_chn,
                                       mid_channels=intermediate_chn, out_channels=out_ch, depth=rescale_module_depth)

    def forward(self, x):
        x = self.encoder(x)
        x = self.rescaler(x)
        return x


class MergedRescaleDecoder(nn.Module):
    def __init__(self, z_channels, out_ch, resolution, num_res_blocks, attn_resolutions, ch, ch_mult=(1,2,4,8),
                 dropout=0.0, resamp_with_conv=True, rescale_factor=1.0, rescale_module_depth=1):
        super().__init__()
        tmp_chn = z_channels*ch_mult[-1]
        self.decoder = Decoder(out_ch=out_ch, z_channels=tmp_chn, attn_resolutions=attn_resolutions, dropout=dropout,
                               resamp_with_conv=resamp_with_conv, in_channels=None, num_res_blocks=num_res_blocks,
                               ch_mult=ch_mult, resolution=resolution, ch=ch)
        self.rescaler = LatentRescaler(factor=rescale_factor, in_channels=z_channels, mid_channels=tmp_chn,
                                       out_channels=tmp_chn, depth=rescale_module_depth)

    def forward(self, x):
        x = self.rescaler(x)
        x = self.decoder(x)
        return x


class Upsampler(nn.Module):
    def __init__(self, in_size, out_size, in_channels, out_channels, ch_mult=2):
        super().__init__()
        assert out_size >= in_size
        num_blocks = int(np.log2(out_size//in_size))+1
        factor_up = 1.+ (out_size % in_size)
        logger.debug(f"Building {self.__class__.__name__} with in_size: {in_size} --> out_size {out_size} and factor {factor_up}")
        self.rescaler = LatentRescaler(factor=factor_up, in_channels=in_channels, mid_channels=2*in_channels,
                                       out_channels=in_channels)
        self.decoder = Decoder(out_ch=out_channels, resolution=out_size, z_channels=in_channels, num_res_blocks=2,
                               attn_resolutions=[], in_channels=None, ch=in_channels,
                               ch_mult=[ch_mult for _ in range(num_blocks)])

    def forward(self, x):
        x = self.rescaler(x)
        x = self.decoder(x)
        return x


class Resize(nn.Module):
    def __init__(self, in_channels=None, learned=False, mode="bilinear"):
        super().__init__()
        self.with_conv = learned
        self.mode = mode
        if self.with_conv:
            logger.debug(f"Note: {self.__class__.__name} uses learned downsampling and will ignore the fixed {mode} mode")
            raise NotImplementedError()
            assert in_channels is not None
            # no asymmetric padding in torch conv, must do it ourselves
            self.conv = torch.nn.Conv2d(in_channels,
                                        in_channels,
                                        kernel_size=4,
                                        stride=2,
                                        padding=1)

    def forward(self, x, scale_factor=1.0):
        if scale_factor==1.0:
            return x
        else:
            x = torch.nn.functional.interpolate(x, mode=self.mode, align_corners=False, scale_factor=scale_factor)
        return x

class FirstStagePostProcessor(nn.Module):

    def __init__(self, ch_mult:list, in_channels,
                 pretrained_model:nn.Module=None,
                 reshape=False,
                 n_channels=None,
                 dropout=0.,
                 pretrained_config=None):
        super().__init__()
        if pretrained_config is None:
            assert pretrained_model is not None, 'Either "pretrained_model" or "pretrained_config" must not be None'
            self.pretrained_model = pretrained_model
        else:
            assert pretrained_config is not None, 'Either "pretrained_model" or "pretrained_config" must not be None'
            self.instantiate_pretrained(pretrained_config)

        self.do_reshape = reshape

        if n_channels is None:
            n_channels = self.pretrained_model.encoder.ch

        self.proj_norm = Normalize(in_channels,num_groups=in_channels//2)
        self.proj = nn.Conv2d(in_channels,n_channels,kernel_size=3,
                            stride=1,padding=1)

        blocks = []
        downs = []
        ch_in = n_channels
        for m in ch_mult:
            blocks.append(ResnetBlock(in_channels=ch_in,out_channels=m*n_channels,dropout=dropout))
            ch_in = m * n_channels
            downs.append(Downsample(ch_in, with_conv=False))

        self.model = nn.ModuleList(blocks)
        self.downsampler = nn.ModuleList(downs)


    def instantiate_pretrained(self, config):
        model = instantiate_from_config(config)
        self.pretrained_model = model.eval()
        # self.pretrained_model.train = False
        for param in self.pretrained_model.parameters():
            param.requires_grad = False


    @torch.no_grad()
    def encode_with_pretrained(self,x):
        c = self.pretrained_model.encode(x)
        if isinstance(c, DiagonalGaussianDistribution):
            c = c.mode()
        return  c

    def forward(self,x):
        z_fs = self.encode_with_pretrained(x)
        z = self.proj_norm(z_fs)
        z = self.proj(z)
        z = nonlinearity(z)

        for submodel, downmodel in zip(self.model,self.downsampler):
            z = submodel(z,temb=None)
            z = downmodel(z)

        if self.do_reshape:
            z = rearrange(z,'b c h w -> b (h w) c')
        return z




import einops
import k_diffusion as K
import torch
import torch.nn as nn
import warnings
from nataili import disable_progress
from nataili.stable_diffusion.prompt_weights import fix_mismatched_tensors
from nataili.util.cast import autocast_cuda
from nataili.util.logger import logger

warnings.filterwarnings("ignore")


class KDiffusionSampler:
    def __init__(self, m, sampler, v: bool = False, callback=None):
        self.model = m
        self.model_wrap = K.external.CompVisVDenoiser(m) if v else K.external.CompVisDenoiser(m)
        self.schedule = sampler
        self.generation_callback = callback

    def get_sampler_name(self):
        return self.schedule

    @autocast_cuda
    def sample_img2img(
        self,
        init_data,
        S,
        t_enc,
        obliterate,
        conditioning,
        unconditional_guidance_scale,
        unconditional_conditioning,
        x_T,
        karras=False,
        sigma_override: dict = None,
        extra_args=None,
    ):
        logger.debug(f"model.device: {self.model.device}")
        x0, z_mask = init_data
        if sigma_override:
            if "min" not in sigma_override:
                raise ValueError("sigma_override must have a 'min' key")
            if "max" not in sigma_override:
                raise ValueError("sigma_override must have a 'max' key")
            if "rho" not in sigma_override:
                raise ValueError("sigma_override must have a 'rho' key")
        sigma_min = self.model_wrap.sigmas[0] if sigma_override is None else sigma_override["min"]
        sigma_max = self.model_wrap.sigmas[-1] if sigma_override is None else sigma_override["max"]
        sigmas = None
        if karras:
            if sigma_override is None:
                if S > 8:
                    sigmas = K.sampling.get_sigmas_karras(S, 0.0292, 14.6146, 7.0, self.model.device)
                elif S == 8:
                    sigmas = K.sampling.get_sigmas_karras(S, 0.0936, 14.6146, 7.0, self.model.device)
                elif S <= 7 and S > 5:
                    sigmas = K.sampling.get_sigmas_karras(S, 0.1072, 14.6146, 7.0, self.model.device)
                elif S <= 5:
                    sigmas = K.sampling.get_sigmas_karras(S, 0.1072, 7.0796, 9.0, self.model.device)
            else:
                sigmas = K.sampling.get_sigmas_karras(
                    S, sigma_override["min"], sigma_override["max"], sigma_override["rho"], self.model.device
                )
        else:
            sigmas = self.model_wrap.get_sigmas(S)
        x = x_T * sigmas[S - t_enc - 1]
        xi = x0 + x
        if z_mask is not None and obliterate:
            random = torch.randn(z_mask.shape, device=xi.device)
            xi = (z_mask * x) + ((1 - z_mask) * xi)  # NOTE: random is not used here. Check if this is correct.
        if extra_args is None:
            model_wrap_cfg = CFGMaskedDenoiser(self.model_wrap)
        else:
            model_wrap_cfg = CFGPix2PixDenoiser(self.model_wrap)
        if extra_args is None:
            if conditioning.shape[1] != unconditional_conditioning.shape[1]:
                conditioning, unconditional_conditioning = fix_mismatched_tensors(
                    conditioning, unconditional_conditioning, self.model
                )
            extra_args = {
                "cond": conditioning,
                "uncond": unconditional_conditioning,
                "cond_scale": unconditional_guidance_scale,
                "mask": z_mask,
                "x0": x0,
                "xi": xi,
            }
        extra_args["cond"] = extra_args["cond"].to(self.model.device)
        extra_args["uncond"] = extra_args["uncond"].to(self.model.device)
        if extra_args["mask"] is not None:
            extra_args["mask"] = extra_args["mask"].to(self.model.device)
        extra_args["x0"] = extra_args["x0"].to(self.model.device)
        extra_args["xi"] = extra_args["xi"].to(self.model.device)
        sigmas = sigmas[S - t_enc - 1 :]

        samples_ddim = None
        if self.schedule == "dpm_fast":
            samples_ddim = K.sampling.__dict__[f"sample_{self.schedule}"](
                model_wrap_cfg,
                xi,
                sigma_min,
                sigmas[0],
                S,
                extra_args=extra_args,
                disable=disable_progress.active,
                callback=self.generation_callback,
            )
        elif self.schedule == "dpm_adaptive":
            samples_ddim = K.sampling.__dict__[f"sample_{self.schedule}"](
                model_wrap_cfg,
                xi,
                sigma_min,
                sigmas[0],
                extra_args=extra_args,
                disable=disable_progress.active,
                callback=self.generation_callback,
            )
        else:
            samples_ddim = K.sampling.__dict__[f"sample_{self.schedule}"](
                model_wrap_cfg,
                xi,
                sigmas,
                extra_args=extra_args,
                disable=disable_progress.active,
                callback=self.generation_callback,
            )
        #
        return samples_ddim, None

    @autocast_cuda
    def sample(
        self,
        S,
        conditioning,
        unconditional_guidance_scale,
        unconditional_conditioning,
        x_T,
        karras=False,
        sigma_override: dict = None,
        extra_args=None,
    ):
        logger.debug(f"model.device: {self.model.device}")
        if sigma_override:
            if "min" not in sigma_override:
                raise ValueError("sigma_override must have a 'min' key")
            if "max" not in sigma_override:
                raise ValueError("sigma_override must have a 'max' key")
            if "rho" not in sigma_override:
                raise ValueError("sigma_override must have a 'rho' key")
        sigma_min = self.model_wrap.sigmas[0] if sigma_override is None else sigma_override["min"]
        sigma_max = self.model_wrap.sigmas[-1] if sigma_override is None else sigma_override["max"]
        sigmas = None
        if karras:
            if sigma_override is None:
                if S > 8:
                    sigmas = K.sampling.get_sigmas_karras(S, 0.0292, 14.6146, 7.0, self.model.device)
                elif S == 8:
                    sigmas = K.sampling.get_sigmas_karras(S, 0.0936, 14.6146, 7.0, self.model.device)
                elif S <= 7 and S > 5:
                    sigmas = K.sampling.get_sigmas_karras(S, 0.1072, 14.6146, 7.0, self.model.device)
                elif S <= 5:
                    sigmas = K.sampling.get_sigmas_karras(S, 0.1072, 7.0796, 9.0, self.model.device)
            else:
                sigmas = K.sampling.get_sigmas_karras(
                    S, sigma_override["min"], sigma_override["max"], sigma_override["rho"], self.model.device
                )
        else:
            sigmas = self.model_wrap.get_sigmas(S)
        x = x_T * sigmas[0]

        if extra_args is None:
            model_wrap_cfg = CFGDenoiser(self.model_wrap)
        else:
            model_wrap_cfg = CFGPix2PixDenoiser(self.model_wrap)
        if extra_args is None:
            if conditioning.shape[1] != unconditional_conditioning.shape[1]:
                conditioning, unconditional_conditioning = fix_mismatched_tensors(
                    conditioning, unconditional_conditioning, self.model
                )
            extra_args = {
                "cond": conditioning,
                "uncond": unconditional_conditioning,
                "cond_scale": unconditional_guidance_scale,
            }
            extra_args["cond"] = extra_args["cond"].to(self.model.device)
            extra_args["uncond"] = extra_args["uncond"].to(self.model.device)
        else:
            # pix2pix, assume it is correct
            pass
        samples_ddim = None
        if self.schedule == "dpm_fast":
            samples_ddim = K.sampling.__dict__[f"sample_{self.schedule}"](
                model_wrap_cfg,
                x,
                sigma_min,
                sigma_max,
                S,
                extra_args=extra_args,
                disable=disable_progress.active,
                callback=self.generation_callback,
            )
        elif self.schedule == "dpm_adaptive":
            samples_ddim = K.sampling.__dict__[f"sample_{self.schedule}"](
                model_wrap_cfg,
                x,
                sigma_min,
                sigma_max,
                extra_args=extra_args,
                disable=disable_progress.active,
                callback=self.generation_callback,
            )
        else:
            samples_ddim = K.sampling.__dict__[f"sample_{self.schedule}"](
                model_wrap_cfg,
                x,
                sigmas,
                extra_args=extra_args,
                disable=disable_progress.active,
                callback=self.generation_callback,
            )
        #
        return samples_ddim, None


class CFGMaskedDenoiser(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.inner_model = model

    def forward(self, x, sigma, uncond, cond, cond_scale, mask, x0, xi):
        x_in = x
        x_in = torch.cat([x_in] * 2)
        sigma_in = torch.cat([sigma] * 2)
        cond_in = torch.cat([uncond, cond])
        uncond, cond = self.inner_model(x_in, sigma_in, cond=cond_in).chunk(2)
        denoised = uncond + (cond - uncond) * cond_scale

        if mask is not None:
            assert x0 is not None
            img_orig = x0
            mask_inv = 1.0 - mask
            denoised = (img_orig * mask_inv) + (mask * denoised)

        return denoised


class CFGDenoiser(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.inner_model = model

    def forward(self, x, sigma, uncond, cond, cond_scale):
        x_in = torch.cat([x] * 2)
        sigma_in = torch.cat([sigma] * 2)
        cond_in = torch.cat([uncond, cond])
        uncond, cond = self.inner_model(x_in, sigma_in, cond=cond_in).chunk(2)
        return uncond + (cond - uncond) * cond_scale


class CFGPix2PixDenoiser(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.inner_model = model

    def forward(self, z, sigma, cond, uncond, text_cfg_scale, image_cfg_scale, mask, x0):
        cfg_z = einops.repeat(z, "1 ... -> n ...", n=3)
        cfg_sigma = einops.repeat(sigma, "1 ... -> n ...", n=3)
        cfg_cond = {
            "c_crossattn": [torch.cat([cond["c_crossattn"][0], uncond["c_crossattn"][0], uncond["c_crossattn"][0]])],
            "c_concat": [torch.cat([cond["c_concat"][0], cond["c_concat"][0], uncond["c_concat"][0]])],
        }
        out_cond, out_img_cond, out_uncond = self.inner_model(cfg_z, cfg_sigma, cond=cfg_cond).chunk(3)
        denoised = (
            out_uncond + text_cfg_scale * (out_cond - out_img_cond) + image_cfg_scale * (out_img_cond - out_uncond)
        )

        if mask is not None:
            assert x0 is not None
            img_orig = x0
            mask_inv = 1.0 - mask
            denoised = (img_orig * mask_inv) + (mask * denoised)

        return denoised



import gradio as gr
from PIL import Image

from nataili.codeformers import codeformers
from nataili.model_manager.codeformer import CodeFormerModelManager
from nataili.ui.gradio.base import GradioUI
from nataili.util.logger import logger


class CodeformerUI(GradioUI):
    def __init__(self):
        model_manager_class = CodeFormerModelManager
        model_name = "CodeFormers"
        inputs = gr.Image(label="Input", type="pil", image_mode="RGB", source="upload")
        outputs = "image"
        title = "Nataili - Face fixing and super resolution with CodeFormers"
        description = "Face fixing and super resolution with CodeFormers"
        thumbnail = ""
        super().__init__(
            model_manager_class=model_manager_class,
            model_name=model_name,
            inputs=inputs,
            outputs=outputs,
            title=title,
            description=description,
            thumbnail=thumbnail,
        )
        self.codeformer = None

    def init(self):
        if self.codeformer is None:
            self.codeformer = codeformers(self.model_manager.loaded_models[self.model_name])

    def predict(self, image: Image.Image):
        self.init()
        return self.codeformer(input_image=image)

    def __call__(self):
        self.launch(self.predict)


def main():
    codeformer_ui = CodeformerUI()
    codeformer_ui()


if __name__ == "__main__":
    main()
